

#include "vulkan_setup.h"

#define VK_USE_PLATFORM_WIN32_KHR
#include <vulkan/vulkan.h>


#include <iostream>
#include <vector>

#include <fstream>


#define VERBOSE_VULKAN_QUEUE_INFO
#define VERBOSE_VULKAN_MEMORY_INFO


//#define VULKAN_INLCUDE_HOST_INPUT_STAGE
//#define VULKAN_INLCUDE_BARRIER_STAGE
//#define VULKAN_INCLUDE_DUMMY_DRAW_STAGE



using namespace std;


typedef PFN_vkVoidFunction(VKAPI_PTR* PFN_vkGetInstanceProcAddr)(VkInstance instance, const char* pName);




//#pragma comment(lib, "vulkan-1")


// maximum number of attempts to ackquire swap image
static constexpr const int SWAP_IMAGE_TIMEOUT = 16;

#define VULKAN_IMAGE_WIDTH	1024
#define VULKAN_IMAGE_HEIGHT	1024

#define NUMBER_OF_XWORKGROUPS	1

// the size in bytes of the scratch space for compute shader input (start/end index,workspace synchronization init, CLUT, VARS, etc)
// this should always be the same
#define SCRATCH_SPACE_SIZE_BYTES	(16384)


// number of swap images
#define VULKAN_SWAP_IMAGE_COUNT		(3)


static constexpr uint32_t DEFAULT_NUM_XWORKGROUPS = 1;


uint32_t vulkan_loaded = false;


// need to know what commands are currently recorded (compute+drawscreen or drawscreen-only)
enum { VULKAN_COMMANDS_RECORDED_NONE = 0, VULKAN_COMMANDS_RECORDED_COMPUTE_DRAW, VULKAN_COMMANDS_RECORDED_DRAW_ONLY };
int vulkan_commands_recorded = VULKAN_COMMANDS_RECORDED_NONE;


VkAllocationCallbacks* vulkan_pAllocator = nullptr;

uint32_t vulkan_alloc_count;
uint32_t vulkan_alloc_size;

uint32_t vulkan_new_job_submit;

// size of the screen/window being output to in pixels
uint32_t vulkan_screen_size_x;
uint32_t vulkan_screen_size_y;

// size of the buffer holding the screen data in 32-bit pixels
uint32_t vulkan_buffer_size_x;
uint32_t vulkan_buffer_size_y;

uint32_t vulkan_num_gpu_x_workgroups = DEFAULT_NUM_XWORKGROUPS;


uint32_t vulkan_SubgroupSize;
uint32_t vulkan_NumSubgroups;
uint32_t vulkan_WorkgroupSize;
uint32_t vulkan_NumWorkgroups;

uint32_t vulkan_image_count;

int32_t* vulkan_mappedmemory;

VkInstance vulkan_instance;

#ifdef NDEBUG
static constexpr const bool enableValidationLayers = false;
#else
static constexpr const bool enableValidationLayers = true;
#endif

VkDebugUtilsMessengerEXT debugMessenger;

VkPhysicalDevice vulkan_physicalDevice;

VkDevice vulkan_device;

VkDeviceMemory vulkan_memory;
VkDeviceMemory vulkan_memory_local;
VkDeviceMemory vulkan_memory_image;

uint32_t vulkan_buffers_count;
//std::vector<VkBuffer> vulkan_buffers(0);
VkBuffer vulkan_buffers[32];
VkDeviceSize vulkan_bufferSizes[32];

uint32_t vulkan_external_vars_count;


uint32_t swap_chain_images_count;
//std::vector<VkImage> swap_chain_images(0);
VkImage swap_chain_images[32];

uint32_t vulkan_surface_formats_count;
//std::vector<VkSurfaceFormatKHR> vulkan_surface_formats(0);
VkSurfaceFormatKHR vulkan_surface_formats[1024];

uint32_t vulkan_present_modes_count;
//std::vector<VkPresentModeKHR> vulkan_present_modes(0);
VkPresentModeKHR vulkan_present_modes[1024];

// device type selected - integrated or discrete
VkPhysicalDeviceType vulkan_device_type;


VkShaderModule vulkan_shader_module;

VkDescriptorSetLayout vulkan_descriptorSetLayout;

VkPipelineLayout vulkan_pipelineLayout;
VkPipeline vulkan_pipeline;

VkCommandPool vulkan_commandPool;
VkCommandBuffer vulkan_commandBuffer;
VkCommandBuffer vulkan_commandBuffer1;
VkQueue vulkan_queue;

VkSubmitInfo vulkan_submitInfo;

uint32_t vulkan_queueFamilyIndex;

VkSemaphore vulkan_image_available_semaphore;
VkSemaphore vulkan_rendering_finished_semaphore;

VkSurfaceKHR vulkan_presentation_surface = nullptr;

VkSwapchainKHR vulkan_swapchain = nullptr;


// command buffers //

// will need to resize this later
uint32_t vulkan_PresentQueueCmdBuffers_count;
//std::vector<VkCommandBuffer> vulkan_PresentQueueCmdBuffers(0);
VkCommandBuffer vulkan_PresentQueueCmdBuffers[32];

// command buffers for global sync 0,250,500,1000 with draw to screen
VkCommandBuffer vulkan_draw_buffer_0[VULKAN_SWAP_IMAGE_COUNT];
VkCommandBuffer vulkan_draw_buffer_256[VULKAN_SWAP_IMAGE_COUNT];
VkCommandBuffer vulkan_draw_buffer_512[VULKAN_SWAP_IMAGE_COUNT];
VkCommandBuffer vulkan_draw_buffer_1024[VULKAN_SWAP_IMAGE_COUNT];
VkCommandBuffer vulkan_draw_buffer_2048[VULKAN_SWAP_IMAGE_COUNT];
VkCommandBuffer vulkan_draw_buffer_4096[VULKAN_SWAP_IMAGE_COUNT];

// command buffers for global sync 0,500,1000 with no draw to screen
VkCommandBuffer vulkan_none_buffer_0;
VkCommandBuffer vulkan_none_buffer_256;
VkCommandBuffer vulkan_none_buffer_512;
VkCommandBuffer vulkan_none_buffer_1024;

// command buffer with only the commands to draw from the pixel buffer to the screen
//uint32_t vulkan_PresentQueueDrawOnlyBuffers_count;
//VkCommandBuffer vulkan_PresentQueueDrawOnlyBuffers[32];




VkDescriptorPool vulkan_descriptorPool;

// this is the image that the compute shader draws the screen into, before the blit to the program window
// now it's going to do a copy which is faster, then the blit to match the size of the screen
VkImage vulkan_image;

VkDescriptorSet vulkan_descriptorSet;


VkFence vulkan_fence;


HMODULE VulkanLibrary;

bool vulkan_load_library()
{
#if defined(VK_USE_PLATFORM_WIN32_KHR)
	VulkanLibrary = LoadLibrary("vulkan-1.dll");
#elif defined(VK_USE_PLATFORM_XCB_KHR) || defined(VK_USE_PLATFORM_XLIB_KHR)
	VulkanLibrary = dlopen("libvulkan.so", RTLD_NOW);
#endif

	if (VulkanLibrary == nullptr) {
		std::cout << "Could not load Vulkan library!" << std::endl;
		return false;
	}
	return true;
}


namespace vulkanapi {

#define VK_EXPORTED_FUNCTION( fun ) PFN_##fun fun;
#define VK_GLOBAL_LEVEL_FUNCTION( fun ) PFN_##fun fun;
#define VK_INSTANCE_LEVEL_FUNCTION( fun ) PFN_##fun fun;
#define VK_DEVICE_LEVEL_FUNCTION( fun ) PFN_##fun fun;

#include "vulkan_list_of_functions.inl"


	bool vulkan_load_exported_entry_points()
	{
#if defined(VK_USE_PLATFORM_WIN32_KHR)
#define LoadProcAddress GetProcAddress
#elif defined(VK_USE_PLATFORM_XCB_KHR) || defined(VK_USE_PLATFORM_XLIB_KHR)
#define LoadProcAddress dlsym
#endif

#define VK_EXPORTED_FUNCTION( fun ) \
if( !(fun = (PFN_##fun)LoadProcAddress( VulkanLibrary, #fun )) ) { \
  std::cout << "Could not load exported function: " << #fun << "!" << std::endl; \
  return false; \
}

#include "vulkan_list_of_functions.inl"

		return true;
	}


	bool vulkan_load_global_level_entry_points()
	{
#define VK_GLOBAL_LEVEL_FUNCTION( fun )                                               \
if( !(fun = (PFN_##fun)vkGetInstanceProcAddr( nullptr, #fun )) ) {                    \
  std::cout << "Could not load global level function: " << #fun << "!" << std::endl;  \
  return false;                                                                       \
}

#include "vulkan_list_of_functions.inl"

		return true;
	}


	bool vulkan_load_instance_level_entry_points()
	{
#define VK_INSTANCE_LEVEL_FUNCTION( fun )                                               \
if( !(fun = (PFN_##fun)vkGetInstanceProcAddr( vulkan_instance, #fun )) ) {              \
  std::cout << "Could not load instance level function: " << #fun << "!" << std::endl;  \
  return false;                                                                         \
}

#include "vulkan_list_of_functions.inl"

		return true;
	}


	bool vulkan_load_device_level_entry_points()
	{
#define VK_DEVICE_LEVEL_FUNCTION( fun )                                               \
if( !(fun = (PFN_##fun)vkGetDeviceProcAddr( vulkan_device, #fun )) ) {                \
  std::cout << "Could not load device level function: " << #fun << "!" << std::endl;  \
  return false;                                                                       \
}

#include "vulkan_list_of_functions.inl"

		return true;
	}

}




const char* VkResultToCString(VkResult result)
{
	switch (result)
	{
	case VK_SUCCESS: return "VK_SUCCESS";
	case VK_NOT_READY: return "VK_NOT_READY";
	case VK_TIMEOUT: return "VK_TIMEOUT";
	case VK_EVENT_SET: return "VK_EVENT_SET";
	case VK_EVENT_RESET: return "VK_EVENT_RESET";
	case VK_INCOMPLETE: return "VK_INCOMPLETE";
	case VK_ERROR_OUT_OF_HOST_MEMORY: return "VK_ERROR_OUT_OF_HOST_MEMORY";
	case VK_ERROR_OUT_OF_DEVICE_MEMORY: return "VK_ERROR_OUT_OF_DEVICE_MEMORY";
	case VK_ERROR_INITIALIZATION_FAILED: return "VK_ERROR_INITIALIZATION_FAILED";
	case VK_ERROR_DEVICE_LOST: return "VK_ERROR_DEVICE_LOST";
	case VK_ERROR_MEMORY_MAP_FAILED: return "VK_ERROR_MEMORY_MAP_FAILED";
	case VK_ERROR_LAYER_NOT_PRESENT: return "VK_ERROR_LAYER_NOT_PRESENT";
	case VK_ERROR_EXTENSION_NOT_PRESENT: return "VK_ERROR_EXTENSION_NOT_PRESENT";
	case VK_ERROR_FEATURE_NOT_PRESENT: return "VK_ERROR_FEATURE_NOT_PRESENT";
	case VK_ERROR_INCOMPATIBLE_DRIVER: return "VK_ERROR_INCOMPATIBLE_DRIVER";
	case VK_ERROR_TOO_MANY_OBJECTS: return "VK_ERROR_TOO_MANY_OBJECTS";
	case VK_ERROR_FORMAT_NOT_SUPPORTED: return "VK_ERROR_FORMAT_NOT_SUPPORTED";
	case VK_ERROR_FRAGMENTED_POOL: return "VK_ERROR_FRAGMENTED_POOL";
	case VK_ERROR_UNKNOWN: return "VK_ERROR_UNKNOWN";
	case VK_ERROR_OUT_OF_POOL_MEMORY: return "VK_ERROR_OUT_OF_POOL_MEMORY";
	case VK_ERROR_INVALID_EXTERNAL_HANDLE: return "VK_ERROR_INVALID_EXTERNAL_HANDLE";
	case VK_ERROR_FRAGMENTATION: return "VK_ERROR_FRAGMENTATION";
	case VK_ERROR_INVALID_OPAQUE_CAPTURE_ADDRESS: return "VK_ERROR_INVALID_OPAQUE_CAPTURE_ADDRESS";
	case VK_PIPELINE_COMPILE_REQUIRED: return "VK_PIPELINE_COMPILE_REQUIRED";
	case VK_ERROR_SURFACE_LOST_KHR: return "VK_ERROR_SURFACE_LOST_KHR";
	case VK_ERROR_NATIVE_WINDOW_IN_USE_KHR: return "VK_ERROR_NATIVE_WINDOW_IN_USE_KHR";
	case VK_SUBOPTIMAL_KHR: return "VK_SUBOPTIMAL_KHR";
	case VK_ERROR_OUT_OF_DATE_KHR: return "VK_ERROR_OUT_OF_DATE_KHR";
	case VK_ERROR_INCOMPATIBLE_DISPLAY_KHR: return "VK_ERROR_INCOMPATIBLE_DISPLAY_KHR";
	case VK_ERROR_VALIDATION_FAILED_EXT: return "VK_ERROR_VALIDATION_FAILED_EXT";
	case VK_ERROR_INVALID_SHADER_NV: return "VK_ERROR_INVALID_SHADER_NV";
#ifdef VK_ENABLE_BETA_EXTENSIONS
	case VK_ERROR_IMAGE_USAGE_NOT_SUPPORTED_KHR: return "VK_ERROR_IMAGE_USAGE_NOT_SUPPORTED_KHR";
#endif
#ifdef VK_ENABLE_BETA_EXTENSIONS
	case VK_ERROR_VIDEO_PICTURE_LAYOUT_NOT_SUPPORTED_KHR: return "VK_ERROR_VIDEO_PICTURE_LAYOUT_NOT_SUPPORTED_KHR";
#endif
#ifdef VK_ENABLE_BETA_EXTENSIONS
	case VK_ERROR_VIDEO_PROFILE_OPERATION_NOT_SUPPORTED_KHR: return "VK_ERROR_VIDEO_PROFILE_OPERATION_NOT_SUPPORTED_KHR";
#endif
#ifdef VK_ENABLE_BETA_EXTENSIONS
	case VK_ERROR_VIDEO_PROFILE_FORMAT_NOT_SUPPORTED_KHR: return "VK_ERROR_VIDEO_PROFILE_FORMAT_NOT_SUPPORTED_KHR";
#endif
#ifdef VK_ENABLE_BETA_EXTENSIONS
	case VK_ERROR_VIDEO_PROFILE_CODEC_NOT_SUPPORTED_KHR: return "VK_ERROR_VIDEO_PROFILE_CODEC_NOT_SUPPORTED_KHR";
#endif
#ifdef VK_ENABLE_BETA_EXTENSIONS
	case VK_ERROR_VIDEO_STD_VERSION_NOT_SUPPORTED_KHR: return "VK_ERROR_VIDEO_STD_VERSION_NOT_SUPPORTED_KHR";
#endif
	case VK_ERROR_INVALID_DRM_FORMAT_MODIFIER_PLANE_LAYOUT_EXT: return "VK_ERROR_INVALID_DRM_FORMAT_MODIFIER_PLANE_LAYOUT_EXT";
	case VK_ERROR_NOT_PERMITTED_KHR: return "VK_ERROR_NOT_PERMITTED_KHR";
	case VK_ERROR_FULL_SCREEN_EXCLUSIVE_MODE_LOST_EXT: return "VK_ERROR_FULL_SCREEN_EXCLUSIVE_MODE_LOST_EXT";
	case VK_THREAD_IDLE_KHR: return "VK_THREAD_IDLE_KHR";
	case VK_THREAD_DONE_KHR: return "VK_THREAD_DONE_KHR";
	case VK_OPERATION_DEFERRED_KHR: return "VK_OPERATION_DEFERRED_KHR";
	case VK_OPERATION_NOT_DEFERRED_KHR: return "VK_OPERATION_NOT_DEFERRED_KHR";
	case VK_ERROR_COMPRESSION_EXHAUSTED_EXT: return "VK_ERROR_COMPRESSION_EXHAUSTED_EXT";
	//case VK_ERROR_OUT_OF_POOL_MEMORY_KHR: return "VK_ERROR_OUT_OF_POOL_MEMORY_KHR";
	//case VK_ERROR_INVALID_EXTERNAL_HANDLE_KHR: return "VK_ERROR_INVALID_EXTERNAL_HANDLE_KHR";
	//case VK_ERROR_FRAGMENTATION_EXT: return "VK_ERROR_FRAGMENTATION_EXT";
	//case VK_ERROR_NOT_PERMITTED_EXT: return "VK_ERROR_NOT_PERMITTED_EXT";
	//case VK_ERROR_INVALID_DEVICE_ADDRESS_EXT: return "VK_ERROR_INVALID_DEVICE_ADDRESS_EXT";
	//case VK_ERROR_INVALID_OPAQUE_CAPTURE_ADDRESS_KHR: return "VK_ERROR_INVALID_OPAQUE_CAPTURE_ADDRESS_KHR";
	//case VK_PIPELINE_COMPILE_REQUIRED_EXT: return "VK_PIPELINE_COMPILE_REQUIRED_EXT";
	//case VK_ERROR_PIPELINE_COMPILE_REQUIRED_EXT: return "VK_ERROR_PIPELINE_COMPILE_REQUIRED_EXT";
	case VK_RESULT_MAX_ENUM: return "VK_RESULT_MAX_ENUM";
	default: return "??????";
	}
}


/**
 * PFN_vkAllocationFunction implementation
 */
void* allocationFunction(void* pUserData, size_t  size, size_t  alignment, VkSystemAllocationScope allocationScope) {

	printf("pAllocator's allocationFunction: size: %u, alignment: %u, allocationScope: %d",
		size, alignment, allocationScope);
	// the allocation itself - ignore alignment, for while
	void* ptr = malloc(size);//_aligned_malloc(size, alignment);
	memset(ptr, 0, size);
	printf(", return ptr* : 0x%p \n", ptr);

	vulkan_alloc_count++;
	vulkan_alloc_size += size;

	return ptr;
}

/**
 * The PFN_vkFreeFunction implementation
 */
void freeFunction(void* pUserData, void* pMemory) {
	printf("pAllocator's freeFunction: ptr: 0x%p\n",
		pMemory);
	// now, the free operation !    
	free(pMemory);

	vulkan_alloc_count--;
}

/**
 * The PFN_vkReallocationFunction implementation
 */
void* reallocationFunction(void* pUserData, void* pOriginal, size_t  size, size_t  alignment, VkSystemAllocationScope allocationScope) {
	printf("pAllocator's REallocationFunction: size %u, alignment %u, allocationScope %d \n",
		size, alignment, allocationScope);
	return realloc(pOriginal, size);
}

/**
 * PFN_vkInternalAllocationNotification implementation
 */
void internalAllocationNotification(void* pUserData, size_t  size, VkInternalAllocationType allocationType, VkSystemAllocationScope                     allocationScope) {
	printf("pAllocator's internalAllocationNotification: size %uz, alignment %uz, allocationType %uz, allocationScope %s \n",
		size,
		allocationType,
		allocationScope);

}

/**
 * PFN_vkInternalFreeNotification implementation
 **/
void internalFreeNotification(void* pUserData, size_t  size, VkInternalAllocationType  allocationType, VkSystemAllocationScope                     allocationScope) {
	printf("pAllocator's internalFreeNotification: size %uz, alignment %uz, allocationType %d, allocationScope %s \n",
		size, allocationType, allocationScope);
}



/**
 * Create Pallocator
 * @param info - String for tracking Allocator usage
 */
static VkAllocationCallbacks* createPAllocator(const char* info) {
	VkAllocationCallbacks* m_allocator = (VkAllocationCallbacks*)malloc(sizeof(VkAllocationCallbacks));
	memset(m_allocator, 0, sizeof(VkAllocationCallbacks));
	m_allocator->pUserData = (void*)info;
	m_allocator->pfnAllocation = (PFN_vkAllocationFunction)(&allocationFunction);
	m_allocator->pfnReallocation = (PFN_vkReallocationFunction)(&reallocationFunction);
	m_allocator->pfnFree = (PFN_vkFreeFunction)&freeFunction;
	m_allocator->pfnInternalAllocation = (PFN_vkInternalAllocationNotification)&internalAllocationNotification;
	m_allocator->pfnInternalFree = (PFN_vkInternalFreeNotification)&internalFreeNotification;
	// storePAllocator(m_allocator);
	return m_allocator;
}



int vulkan_get_subgroup_size()
{
	return vulkan_SubgroupSize;
}


void vulkan_set_screen_size(int screen_size_x, int screen_size_y)
{
	vulkan_screen_size_x = (uint32_t)screen_size_x;
	vulkan_screen_size_y = (uint32_t)screen_size_y;
}

void vulkan_set_buffer_size(int buffer_size_x, int buffer_size_y)
{
	vulkan_buffer_size_x = (uint32_t)buffer_size_x;
	vulkan_buffer_size_y = (uint32_t)buffer_size_y;
}


bool vulkan_is_screen_size_changed(int screen_size_x, int screen_size_y)
{
	return ((screen_size_x != vulkan_screen_size_x) || (screen_size_y != vulkan_screen_size_y));
}
bool vulkan_is_buffer_size_changed(int buffer_size_x, int buffer_size_y)
{
	return ((buffer_size_x != vulkan_buffer_size_x) || (buffer_size_y != vulkan_buffer_size_y));
}


void vulkan_set_gpu_threads(int num_gpu_threads)
{
	vulkan_num_gpu_x_workgroups = (uint32_t)num_gpu_threads;
}

int vulkan_get_gpu_threads()
{
	return vulkan_num_gpu_x_workgroups;
}


// wait for gpu with vulkan
bool vulkan_wait()
{
	VkResult v1;

	// get device queue ??
	vulkanapi::vkGetDeviceQueue(vulkan_device, vulkan_queueFamilyIndex, 0, &vulkan_queue);

	//VkResult v1 = vulkanapi::vkQueueWaitIdle(vulkan_queue);
	//if ( v1 != VK_SUCCESS)
	if (v1 = vulkanapi::vkQueueWaitIdle(vulkan_queue))
	{
		std::cout << "\nVULKAN ERROR: Problem with idle wait.\n";
		std::cout << "\nError value: " << v1 << ": " << VkResultToCString(v1);
		return false;
	}

	return true;
}


// reset gpu fence
bool vulkan_reset_fence()
{
	VkResult res;

	//VkResult res = vulkanapi::vkResetFences(vulkan_device, 1, &vulkan_fence);
	//if (res != VK_SUCCESS)
	if (res = vulkanapi::vkResetFences(vulkan_device, 1, &vulkan_fence))
	{
		std::cout << "\nVULKAN ERROR: Problem with reset FENCE.\n";
		std::cout << "\nError value: " << res << ": " << VkResultToCString(res);
		return false;
	}

	return true;
}

// wait for gpu fence with vulkan
bool vulkan_wait_fence()
{
	VkResult res;

	//VkResult res = vulkanapi::vkWaitForFences(vulkan_device, 1, &vulkan_fence, VK_TRUE, UINT64_MAX);
	//if (res != VK_SUCCESS)
	if (res = vulkanapi::vkWaitForFences(vulkan_device, 1, &vulkan_fence, VK_TRUE, UINT64_MAX))
	{
		std::cout << "\nVULKAN ERROR: Problem with idle FENCE wait: " << VkResultToCString(res);
		std::cout << "\nError value: " << res;
		return false;
	}

	//vulkanapi::vkResetFences(vulkan_device, 1, &vulkan_fence);
	return true;
}

// flush mapped memory ??
void vulkan_flush()
{
	VkResult res;

	VkMappedMemoryRange memrange = {
		VK_STRUCTURE_TYPE_MAPPED_MEMORY_RANGE,
		nullptr,
		vulkan_memory,
		0,
		VK_WHOLE_SIZE
	};

	//if (vulkanapi::vkFlushMappedMemoryRanges(vulkan_device, 1, &memrange) != VK_SUCCESS)
	if(res= vulkanapi::vkFlushMappedMemoryRanges(vulkan_device, 1, &memrange))
	{
		std::cout << "\nVULKAN: ERROR: problem flushing mapped memory range. Error#" << res << ": " << VkResultToCString(res);
	}
}




// pre-record a generic draw run
// commandBuffer - the vulkan command buffer to record to
// thread_count - number of workgroups to use when invoking the shaders
// global_barrier_count - the number of global "barriers" (local barriers are in the shader code, but not global)
// draw_to_screen - do these draw commands finish with sending the buffer to the screen at the end?
bool vulkan_prerecord_draw_commands(VkCommandBuffer &commandBuffer, int image_index, int thread_count, int global_barrier_count, bool draw_to_screen)
{
	// get the buffer to use for staging, which is going to be the last one
	//VkBuffer vulkan_staging_buffer = vulkan_buffers[vulkan_buffers.size() - 1];
	VkBuffer vulkan_staging_buffer = vulkan_buffers[vulkan_buffers_count - 1];

	// 4MB
	//VkBuffer vulkan_main_framebuffer = vulkan_buffers[1];

	// 16MB
	//VkBuffer vulkan_work_data = vulkan_buffers[8];

	// 4096 bytes
	//VkBuffer vulkan_scratch_data = vulkan_buffers[0];

	// 1MB
	//VkBuffer vulkan_var_data = vulkan_buffers[7];


	VkImageSubresourceRange image_subresource_range = { VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 };


	VkImageBlit image_blit_region = {
		{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1 },
		{ { 0, 0, 0 }, { 640, 480, 1 } },
		{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1 },
		//{ { 0, 0, 0 }, { 640, 480, 1 } }
		{ { 0, 0, 0 }, { (int32_t)vulkan_screen_size_x, (int32_t)vulkan_screen_size_y, 1 } }
	};


	VkBufferImageCopy region{};
	region.bufferOffset = 0;
	region.bufferRowLength = 0;
	region.bufferImageHeight = 0;

	region.imageSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
	region.imageSubresource.mipLevel = 0;
	region.imageSubresource.baseArrayLayer = 0;
	region.imageSubresource.layerCount = 1;

	region.imageOffset = { 0, 0, 0 };
	region.imageExtent = {
		640,	// width
		480,	// height
		1
	};


	// image barrier for the image that compute shader writes to
	VkImageMemoryBarrier barrier_wait_until_image_written[2] =
	{
	{
		VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
		nullptr,                                    // const void                            *pNext
		VK_ACCESS_TRANSFER_WRITE_BIT,                  // VkAccessFlags                          srcAccessMask
		VK_ACCESS_TRANSFER_READ_BIT,               // VkAccessFlags                          dstAccessMask
		VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,                  // VkImageLayout                          oldLayout
		VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,				// VkImageLayout                          newLayout
		vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
		vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
		vulkan_image,                       // VkImage                                image
		image_subresource_range                     // VkImageSubresourceRange                subresourceRange
	},
	{
		VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
		nullptr,                                    // const void                            *pNext
		VK_ACCESS_TRANSFER_READ_BIT,                  // VkAccessFlags                          srcAccessMask
		VK_ACCESS_TRANSFER_WRITE_BIT,               // VkAccessFlags                          dstAccessMask
		VK_IMAGE_LAYOUT_UNDEFINED,                  // VkImageLayout                          oldLayout
		VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,				// VkImageLayout                          newLayout
		vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
		vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
		swap_chain_images[image_index],                       // VkImage                                image
		image_subresource_range                     // VkImageSubresourceRange                subresourceRange
	}
	};


	VkBufferMemoryBarrier barrier_wait_until_buffer_written = {
		VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER,
		nullptr,
		VK_ACCESS_SHADER_WRITE_BIT,
		VK_ACCESS_TRANSFER_READ_BIT,
		vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
		vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
		vulkan_staging_buffer,
		0,
		640 * 480 * 4
	};


	VkImageMemoryBarrier barrier_from_clear_to_present[1] =
	{
	{
		VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
		nullptr,                                    // const void                            *pNext
		VK_ACCESS_TRANSFER_WRITE_BIT,               // VkAccessFlags                          srcAccessMask
		VK_ACCESS_TRANSFER_READ_BIT,                  // VkAccessFlags                          dstAccessMask
		VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,       // VkImageLayout                          oldLayout
		VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,            // VkImageLayout                          newLayout
		vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
		vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
		swap_chain_images[image_index],                       // VkImage                                image
		image_subresource_range                     // VkImageSubresourceRange                subresourceRange
	}

	};

	VkImageMemoryBarrier barrier_from_present_to_clear = {
		VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
		nullptr,                                    // const void                            *pNext
		VK_ACCESS_MEMORY_READ_BIT,                  // VkAccessFlags                          srcAccessMask
		VK_ACCESS_TRANSFER_WRITE_BIT,               // VkAccessFlags                          dstAccessMask
		VK_IMAGE_LAYOUT_UNDEFINED,                  // VkImageLayout                          oldLayout
		VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,       // VkImageLayout                          newLayout
		vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
		vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
		vulkan_image,                       // VkImage                                image
		image_subresource_range                     // VkImageSubresourceRange                subresourceRange
	};


	VkMemoryBarrier barrier_memory_compute_shader = {
		VK_STRUCTURE_TYPE_MEMORY_BARRIER,
		nullptr,
		VK_ACCESS_SHADER_WRITE_BIT,
		VK_ACCESS_SHADER_READ_BIT
	};

	VkMemoryBarrier barrier_memory_compute_totransfer = {
		VK_STRUCTURE_TYPE_MEMORY_BARRIER,
		nullptr,
		VK_ACCESS_SHADER_WRITE_BIT,
		VK_ACCESS_TRANSFER_READ_BIT
	};


	VkBufferCopy bufcopy =
	{
		0, 0, 1024 * 512 * 2
	};

	VkBufferCopy bufcopy_dummy =
	{
		0, 1, 1
	};


	//if (vulkanapi::vkAllocateCommandBuffers(vulkan_device, &commandBufferAllocateInfo, &vulkan_commandBuffer) != VK_SUCCESS)
	//{
	//	std::cout << "\nVULKAN ERROR: Problem allocating command buffers for swap chain commands.\n";
	//	return false;
	//}

	VkCommandBufferAllocateInfo commandBufferAllocateInfo = {
	  VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
	  0,
	  vulkan_commandPool,
	  VK_COMMAND_BUFFER_LEVEL_PRIMARY,
	  1
	};

	if (vulkanapi::vkAllocateCommandBuffers(vulkan_device, &commandBufferAllocateInfo, &commandBuffer) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating command buffer.\n";
		return false;

	}

	// reset command buffer to initial state
	vulkanapi::vkResetCommandBuffer(commandBuffer, 0);


	VkCommandBufferBeginInfo commandBufferBeginInfo = {
	  VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
	  0,
	  VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT,
	  //VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT,
	  0
	};

	vulkanapi::vkBeginCommandBuffer(commandBuffer, &commandBufferBeginInfo);


	// I'd imagine the compute shader run should go here
	vulkanapi::vkCmdBindPipeline(commandBuffer, VK_PIPELINE_BIND_POINT_COMPUTE, vulkan_pipeline);

	vulkanapi::vkCmdBindDescriptorSets(commandBuffer, VK_PIPELINE_BIND_POINT_COMPUTE, vulkan_pipelineLayout, 0, 1, &vulkan_descriptorSet, 0, 0);



	for (int i = 0; i <= global_barrier_count; i++)
	{

		if (i)
		{
			vulkanapi::vkCmdPipelineBarrier(commandBuffer, VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, 0, 1, &barrier_memory_compute_shader, 0, nullptr, 0, nullptr);
		}

		// set number of shaders to run simultaneously here
		vulkanapi::vkCmdDispatch(commandBuffer, thread_count, 1, 1);
	}


	if (draw_to_screen)
	{

		// draw to frame buffer
		vulkanapi::vkCmdPipelineBarrier(commandBuffer, VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 1, &barrier_memory_compute_totransfer, 1, &barrier_wait_until_buffer_written, 0, nullptr);
		vulkanapi::vkCmdPipelineBarrier(commandBuffer, VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, &barrier_from_present_to_clear);



		// draw the screen from buffer
		vulkanapi::vkCmdCopyBufferToImage(commandBuffer, vulkan_staging_buffer, vulkan_image, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &region);

		vulkanapi::vkCmdPipelineBarrier(commandBuffer, VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 2, (const VkImageMemoryBarrier*)&barrier_wait_until_image_written);

		// blit image to the screen
		vulkanapi::vkCmdBlitImage(commandBuffer, vulkan_image, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, swap_chain_images[image_index], VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &image_blit_region, VK_FILTER_NEAREST);


		// wait for all previous commands to finish
		vulkanapi::vkCmdPipelineBarrier(commandBuffer, VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, (const VkImageMemoryBarrier*)&barrier_from_clear_to_present);
	}

	if (vulkanapi::vkEndCommandBuffer(commandBuffer) != VK_SUCCESS) {
		std::cout << "ERROR: VULKAN: Could not record command buffers!" << std::endl;
		return false;
	}

	return true;
}



bool vulkan_prerecord_command_buffers(int max_thread_count)
{
	// clear out any previous recordings
	//vulkan_clear();

	cout << "\nVULKAN: INFO: Recording commands with max number of threads: " << max_thread_count;

	// record a command buffer that draws to the screen for each swap image
	for (int i = 0; i < VULKAN_SWAP_IMAGE_COUNT; i++)
	{
		// single threaded without global synchronization
		vulkan_prerecord_draw_commands(vulkan_draw_buffer_0[i], i, 1, 0, true);

		// single OR multi-threaded with global synchronization
		vulkan_prerecord_draw_commands(vulkan_draw_buffer_256[i], i, max_thread_count, 256, true);
		vulkan_prerecord_draw_commands(vulkan_draw_buffer_512[i], i, max_thread_count, 512, true);
		vulkan_prerecord_draw_commands(vulkan_draw_buffer_1024[i], i, max_thread_count, 1024, true);
		vulkan_prerecord_draw_commands(vulkan_draw_buffer_2048[i], i, max_thread_count, 2048, true);
		vulkan_prerecord_draw_commands(vulkan_draw_buffer_4096[i], i, max_thread_count, 4096, true);
	}

	// record the command buffers that don't draw to screen
	vulkan_prerecord_draw_commands(vulkan_none_buffer_0, 0, 1, 0, false);
	vulkan_prerecord_draw_commands(vulkan_none_buffer_256, 0, max_thread_count, 256, false);
	vulkan_prerecord_draw_commands(vulkan_none_buffer_512, 0, max_thread_count, 512, false);
	vulkan_prerecord_draw_commands(vulkan_none_buffer_1024, 0, max_thread_count, 1024, false);

	return true;
}




bool vulkan_execute_draw_only()
{
	//if (vkQueueWaitIdle(vulkan_queue) != VK_SUCCESS)
	//{
	//	cout << "\nERROR: VULKAN: Problem with idle wait during vulkan_execute.\n";
	//}

	uint32_t image_index;

	VkResult result;

	//result = vulkanapi::vkAcquireNextImageKHR(vulkan_device, vulkan_swapchain, UINT64_MAX, vulkan_image_available_semaphore, VK_NULL_HANDLE, &image_index);

	// note: for some reason on some platforms you have to keep trying to ackquire the swap image
	for (int i = 0; i < SWAP_IMAGE_TIMEOUT; i++)
	{
		result = vulkanapi::vkAcquireNextImageKHR(vulkan_device, vulkan_swapchain, UINT64_MAX, vulkan_image_available_semaphore, VK_NULL_HANDLE, &image_index);

		if ((result == VK_SUCCESS) || (result == VK_SUBOPTIMAL_KHR)) break;

		if (result == VK_ERROR_OUT_OF_DATE_KHR)
		{
			std::cout << "\nVULKAN: WHILE ACQUIRING IMAGE IT WAS OUT OF DATE.. POSSIBLE WINDOW RESIZE?\n";

			// also need to re-record command buffers using new swapchain
			vulkan_record_draw_only_commands();
		}
	}



	switch (result) {
	case VK_SUCCESS:
	case VK_SUBOPTIMAL_KHR:
		break;
	case VK_ERROR_OUT_OF_DATE_KHR:
		// TODO: recreate swap stuff
		//return OnWindowSizeChanged();
		std::cout << "\nProblem occurred during swap chain image acquisition!: VK_ERROR_OUT_OF_DATE_KHR" << std::endl;

		// try recreating swap chain ?
		vulkan_create_swap_chain();

		// also need to re-record command buffers using new swapchain
		vulkan_record_draw_only_commands();

		// must do this because the frame still needs to execute even though there is no image index
		//vulkan_execute_compute_only();

		return true;
		break;

	default:

		std::cout << "\nProblem occurred during swap chain image acquisition!:" << result << ": " << VkResultToCString(result);

		// try recreating swap chain ?
		vulkan_create_swap_chain();

		return false;
		break;
	}


	VkPipelineStageFlags wait_dst_stage_mask = VK_PIPELINE_STAGE_TRANSFER_BIT;
	VkSubmitInfo submit_info = {
	  VK_STRUCTURE_TYPE_SUBMIT_INFO,                // VkStructureType              sType
	  nullptr,                                      // const void                  *pNext
	  1,                                            // uint32_t                     waitSemaphoreCount
	  &vulkan_image_available_semaphore,              // const VkSemaphore           *pWaitSemaphores
	  &wait_dst_stage_mask,                         // const VkPipelineStageFlags  *pWaitDstStageMask;
	  1,                                            // uint32_t                     commandBufferCount
	  & vulkan_PresentQueueCmdBuffers[image_index],  // const VkCommandBuffer       *pCommandBuffers
	  1,                                            // uint32_t                     signalSemaphoreCount
	  &vulkan_rendering_finished_semaphore            // const VkSemaphore           *pSignalSemaphores
	};

	vulkanapi::vkResetFences(vulkan_device, 1, &vulkan_fence);

	//if (vulkanapi::vkQueueSubmit(vulkan_queue, 1, &submit_info, VK_NULL_HANDLE) != VK_SUCCESS) {
	if (result = vulkanapi::vkQueueSubmit(vulkan_queue, 1, &submit_info, vulkan_fence))
	{
		cout << "\nVULKAN ERROR: Problem with queue submit. Error#" << dec << result << ": " << VkResultToCString(result);

		return false;
	}


	VkPresentInfoKHR present_info = {
		  VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,           // VkStructureType              sType
		  nullptr,                                      // const void                  *pNext
		  1,                                            // uint32_t                     waitSemaphoreCount
		  &vulkan_rendering_finished_semaphore,           // const VkSemaphore           *pWaitSemaphores
		  1,                                            // uint32_t                     swapchainCount
		  &vulkan_swapchain,                            // const VkSwapchainKHR        *pSwapchains
		  &image_index,                                 // const uint32_t              *pImageIndices
		  nullptr                                       // VkResult                    *pResults
	};
	result = vulkanapi::vkQueuePresentKHR(vulkan_queue, &present_info);

	switch (result) {
	case VK_SUCCESS:
		break;
	case VK_ERROR_OUT_OF_DATE_KHR:
	case VK_SUBOPTIMAL_KHR:
		// TODO: recreate swap stuff
		//return OnWindowSizeChanged();
		std::cout << "\nProblem occurred during image presentation!: VK_ERROR_OUT_OF_DATE_KHR or VK_SUBOPTIMAL_KHR" << std::endl;

		// try recreating swap chain ?
		vulkan_create_swap_chain();

		// also need to re-record command buffers using new swapchain
		vulkan_record_draw_only_commands();

		break;

	default:
		std::cout << "\nProblem occurred during image presentation!:" << result << std::endl;

		// try recreating swap chain ?
		//vulkan_create_swap_chain();

		return false;
		break;
	}

	vulkan_new_job_submit = 1;

	return true;
}


bool vulkan_execute_compute_only()
{
	//if (vkQueueWaitIdle(vulkan_queue) != VK_SUCCESS)
	//{
	//	cout << "\nERROR: VULKAN: Problem with idle wait during vulkan_execute.\n";
	//}

	VkSubmitInfo submit_info = {
	  VK_STRUCTURE_TYPE_SUBMIT_INFO,
	  0,
	  0,
	  0,
	  0,
	  1,
	  //&vulkan_commandBuffer1,
	  & vulkan_none_buffer_0,
	  0,
	  0
	};

	vulkanapi::vkResetFences(vulkan_device, 1, &vulkan_fence);
	//if (vulkanapi::vkQueueSubmit(vulkan_queue, 1, &vulkan_submitInfo, 0) != VK_SUCCESS)
	if (vulkanapi::vkQueueSubmit(vulkan_queue, 1, &submit_info, vulkan_fence) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem with queue submit (COMPUTE ONLY).\n";
		return false;
	}

	vulkan_new_job_submit = 1;

	return true;
}

// submit the command to run the compute job
// can add more commands later as needed, should be same for both ps1 and ps2
bool vulkan_execute()
{
	//if (vkQueueWaitIdle(vulkan_queue) != VK_SUCCESS)
	//{
	//	cout << "\nERROR: VULKAN: Problem with idle wait during vulkan_execute.\n";
	//}

	uint32_t image_index;

	VkResult result;
	
	//result = vulkanapi::vkAcquireNextImageKHR(vulkan_device, vulkan_swapchain, UINT64_MAX, vulkan_image_available_semaphore, VK_NULL_HANDLE, &image_index);

	// note: for some reason on some platforms you have to keep trying to ackquire the swap image
	for (int i = 0; i < SWAP_IMAGE_TIMEOUT; i++)
	{
		result = vulkanapi::vkAcquireNextImageKHR(vulkan_device, vulkan_swapchain, UINT64_MAX, vulkan_image_available_semaphore, VK_NULL_HANDLE, &image_index);

		if ((result == VK_SUCCESS) || (result == VK_SUBOPTIMAL_KHR)) break;

		if (result == VK_ERROR_OUT_OF_DATE_KHR)
		{
			std::cout << "\nVULKAN: WHILE ACQUIRING IMAGE IT WAS OUT OF DATE.. POSSIBLE WINDOW RESIZE?\n";

			// try recreating swap chain ?
			vulkan_create_swap_chain();

			// also need to re-record command buffers using new swapchain
			vulkan_record_default_commands();
		}
	}
	


	switch (result) {
	case VK_SUCCESS:
	case VK_SUBOPTIMAL_KHR:
		break;
	case VK_ERROR_OUT_OF_DATE_KHR:
		// TODO: recreate swap stuff
		//return OnWindowSizeChanged();
		std::cout << "\nProblem occurred during swap chain image acquisition!: VK_ERROR_OUT_OF_DATE_KHR" << std::endl;

		// try recreating swap chain ?
		vulkan_create_swap_chain();

		// also need to re-record command buffers using new swapchain
		vulkan_record_default_commands();

		// must do this because the frame still needs to execute even though there is no image index
		vulkan_execute_compute_only();

		return true;
		break;

	default:

		std::cout << "\nProblem occurred during swap chain image acquisition!:" << result << std::endl;

		// try recreating swap chain ?
		//vulkan_create_swap_chain();

		// must do this because the frame still needs to execute even though there is no image index
		vulkan_execute_compute_only();

		return false;
		break;
	}


	VkPipelineStageFlags wait_dst_stage_mask = VK_PIPELINE_STAGE_TRANSFER_BIT;
	VkSubmitInfo submit_info = {
	  VK_STRUCTURE_TYPE_SUBMIT_INFO,                // VkStructureType              sType
	  nullptr,                                      // const void                  *pNext
	  1,                                            // uint32_t                     waitSemaphoreCount
	  & vulkan_image_available_semaphore,              // const VkSemaphore           *pWaitSemaphores
	  & wait_dst_stage_mask,                         // const VkPipelineStageFlags  *pWaitDstStageMask;
	  1,                                            // uint32_t                     commandBufferCount
	  //& vulkan_PresentQueueCmdBuffers[image_index],  // const VkCommandBuffer       *pCommandBuffers
	  & vulkan_draw_buffer_0[image_index],
	  1,                                            // uint32_t                     signalSemaphoreCount
	  & vulkan_rendering_finished_semaphore            // const VkSemaphore           *pSignalSemaphores
	};

	vulkanapi::vkResetFences(vulkan_device, 1, &vulkan_fence);
	//if (vulkanapi::vkQueueSubmit(vulkan_queue, 1, &submit_info, VK_NULL_HANDLE) != VK_SUCCESS) {
	if (result = vulkanapi::vkQueueSubmit(vulkan_queue, 1, &submit_info, vulkan_fence))
	{
		cout << "\nVULKAN ERROR: Problem with queue submit. Error#" << dec << result << ": " << VkResultToCString(result);
		return false;
	}


	VkPresentInfoKHR present_info = {
		  VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,           // VkStructureType              sType
		  nullptr,                                      // const void                  *pNext
		  1,                                            // uint32_t                     waitSemaphoreCount
		  & vulkan_rendering_finished_semaphore,           // const VkSemaphore           *pWaitSemaphores
		  1,                                            // uint32_t                     swapchainCount
		  & vulkan_swapchain,                            // const VkSwapchainKHR        *pSwapchains
		  & image_index,                                 // const uint32_t              *pImageIndices
		  nullptr                                       // VkResult                    *pResults
	};
	result = vulkanapi::vkQueuePresentKHR(vulkan_queue, &present_info);

	switch (result) {
	case VK_SUCCESS:
		break;
	case VK_ERROR_OUT_OF_DATE_KHR:
	case VK_SUBOPTIMAL_KHR:
		// TODO: recreate swap stuff
		//return OnWindowSizeChanged();
		std::cout << "\nProblem occurred during image presentation!: VK_ERROR_OUT_OF_DATE_KHR or VK_SUBOPTIMAL_KHR" << std::endl;

		// try recreating swap chain ?
		vulkan_create_swap_chain();

		// also need to re-record command buffers using new swapchain
		vulkan_record_default_commands();

		break;

	default:
		std::cout << "\nProblem occurred during image presentation!:" << result << std::endl;

		// try recreating swap chain ?
		vulkan_create_swap_chain();

		return false;
		break;
	}

	vulkan_new_job_submit = 1;

	return true;
}


bool vulkan_execute_display(int global_barrier_count, bool draw_to_screen)
{
	uint32_t image_index;
	VkResult result;

	VkCommandBuffer current_commands_execute;
	VkCommandBuffer current_commands_draw;


	if (draw_to_screen)
	{
		//result = vulkanapi::vkAcquireNextImageKHR(vulkan_device, vulkan_swapchain, UINT64_MAX, vulkan_image_available_semaphore, VK_NULL_HANDLE, &image_index);

		// note: for some reason on some platforms you have to keep trying to ackquire the swap image
		for (int i = 0; i < SWAP_IMAGE_TIMEOUT; i++)
		{
			result = vulkanapi::vkAcquireNextImageKHR(vulkan_device, vulkan_swapchain, UINT64_MAX, vulkan_image_available_semaphore, VK_NULL_HANDLE, &image_index);

			if ((result == VK_SUCCESS) || (result == VK_SUBOPTIMAL_KHR)) break;

			if (result == VK_ERROR_OUT_OF_DATE_KHR)
			{
				std::cout << "\nVULKAN: WHILE ACQUIRING IMAGE IT WAS OUT OF DATE.. POSSIBLE WINDOW RESIZE?\n";

				// try recreating swap chain ?
				vulkan_create_swap_chain();

				// also need to re-record command buffers using new swapchain
				vulkan_record_default_commands();
			}
		}


		switch (result) {
		case VK_SUCCESS:
		case VK_SUBOPTIMAL_KHR:
			break;
		case VK_ERROR_OUT_OF_DATE_KHR:
			// TODO: recreate swap stuff
			//return OnWindowSizeChanged();
			std::cout << "\nProblem occurred during swap chain image acquisition!: VK_ERROR_OUT_OF_DATE_KHR" << std::endl;

			// try recreating swap chain ?
			vulkan_create_swap_chain();

			// also need to re-record command buffers using new swapchain
			vulkan_record_default_commands();

			// must do this because the frame still needs to execute even though there is no image index
			vulkan_execute_compute_only();

			return true;

		default:

			std::cout << "\nProblem occurred during swap chain image acquisition!:" << result << std::endl;

			// must do this because the frame still needs to execute even though there is no image index
			vulkan_execute_compute_only();

			return false;
		}
	}


	// choose the command buffer based on global sync count
	// note: these need to be pre-recorded or else the performance goes downhill
	if (
		global_barrier_count
		&&
		(vulkan_get_gpu_threads() > 1)
		)
	{
		if (global_barrier_count < 256)
		{
			// set the command buffer to use if drawing to screen
			current_commands_draw = vulkan_draw_buffer_256[image_index];

			// set the command buffer to use if won't be drawing results to screen (like pixel read from gpu)
			current_commands_execute = vulkan_none_buffer_256;
		}
		else if (global_barrier_count < 512)
		{
			current_commands_draw = vulkan_draw_buffer_512[image_index];
			current_commands_execute = vulkan_none_buffer_512;
		}
		else if (global_barrier_count < 1024)
		{
			current_commands_draw = vulkan_draw_buffer_1024[image_index];
			current_commands_execute = vulkan_none_buffer_1024;
		}
		else if (global_barrier_count < 2048)
		{
			current_commands_draw = vulkan_draw_buffer_2048[image_index];
			current_commands_execute = vulkan_none_buffer_0;
		}
		else if (global_barrier_count < 4096)
		{
			current_commands_draw = vulkan_draw_buffer_4096[image_index];
			current_commands_execute = vulkan_none_buffer_0;
		}
		else
		{
			// if the number of syncs is too large, then just use a single workgroup with no global syncs
			current_commands_draw = vulkan_draw_buffer_0[image_index];
			current_commands_execute = vulkan_none_buffer_0;
		}

	}
	else
	{
		// zero means to force to single thread and no global sync //
		current_commands_draw = vulkan_draw_buffer_0[image_index];
		current_commands_execute = vulkan_none_buffer_0;
	}


	if (draw_to_screen)
	{
		VkPipelineStageFlags wait_dst_stage_mask = VK_PIPELINE_STAGE_TRANSFER_BIT;
		VkSubmitInfo submit_info = {
		  VK_STRUCTURE_TYPE_SUBMIT_INFO,                // VkStructureType              sType
		  nullptr,                                      // const void                  *pNext
		  1,                                            // uint32_t                     waitSemaphoreCount
		  &vulkan_image_available_semaphore,              // const VkSemaphore           *pWaitSemaphores
		  &wait_dst_stage_mask,                         // const VkPipelineStageFlags  *pWaitDstStageMask;
		  1,                                            // uint32_t                     commandBufferCount
		  //&vulkan_commandBuffer1,  // const VkCommandBuffer       *pCommandBuffers
		  & current_commands_draw,  // const VkCommandBuffer       *pCommandBuffers
		  1,                                            // uint32_t                     signalSemaphoreCount
		  &vulkan_rendering_finished_semaphore            // const VkSemaphore           *pSignalSemaphores
		};

		vulkanapi::vkResetFences(vulkan_device, 1, &vulkan_fence);

		//if (vulkanapi::vkQueueSubmit(vulkan_queue, 1, &submit_info, VK_NULL_HANDLE) != VK_SUCCESS) {
		if (vulkanapi::vkQueueSubmit(vulkan_queue, 1, &submit_info, vulkan_fence) != VK_SUCCESS) {
			//return false;
			cout << "\nVULKAN ERROR: Problem with queue submit.\n";
			return false;
		}


		VkPresentInfoKHR present_info = {
			  VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,           // VkStructureType              sType
			  nullptr,                                      // const void                  *pNext
			  1,                                            // uint32_t                     waitSemaphoreCount
			  &vulkan_rendering_finished_semaphore,           // const VkSemaphore           *pWaitSemaphores
			  1,                                            // uint32_t                     swapchainCount
			  &vulkan_swapchain,                            // const VkSwapchainKHR        *pSwapchains
			  &image_index,                                 // const uint32_t              *pImageIndices
			  nullptr                                       // VkResult                    *pResults
		};

		result = vulkanapi::vkQueuePresentKHR(vulkan_queue, &present_info);


		switch (result) {
		case VK_SUCCESS:
			break;
		case VK_ERROR_OUT_OF_DATE_KHR:
		case VK_SUBOPTIMAL_KHR:
			//return OnWindowSizeChanged();
			std::cout << "Problem occurred during image presentation!: VK_ERROR_OUT_OF_DATE_KHR or VK_SUBOPTIMAL_KHR" << std::endl;

			// try recreating swap chain ?
			vulkan_create_swap_chain();

			// also need to re-record command buffers using new swapchain
			vulkan_record_default_commands();
			break;

		default:
			std::cout << "Problem occurred during image presentation!:" << result << std::endl;

			return false;
			break;
		}
	}
	else
	{
		VkSubmitInfo submit_info = {
		  VK_STRUCTURE_TYPE_SUBMIT_INFO,
		  0,
		  0,
		  0,
		  0,
		  1,
		  //&vulkan_commandBuffer1,
		  & current_commands_execute,
		  0,
		  0
		};

		vulkanapi::vkResetFences(vulkan_device, 1, &vulkan_fence);
		//if (vulkanapi::vkQueueSubmit(vulkan_queue, 1, &submit_info, VK_NULL_HANDLE) != VK_SUCCESS) {
		if (vulkanapi::vkQueueSubmit(vulkan_queue, 1, &submit_info, vulkan_fence) != VK_SUCCESS) {
			//return false;
			cout << "\nVULKAN ERROR: Problem with queue submit.\n";
			return false;
		}
	}

	vulkan_new_job_submit = 1;

	// ran ok
	return true;
}


bool vulkan_record_execute_nodisplay()
{


	return true;
}


// when parameters change need to recreate presentation queue
void vulkan_clear()
{
	// vulkan commands are being cleared
	vulkan_commands_recorded = VULKAN_COMMANDS_RECORDED_NONE;

	if (vulkan_device != VK_NULL_HANDLE)
	{
		vulkanapi::vkDeviceWaitIdle(vulkan_device);

		if (vulkan_commandPool != VK_NULL_HANDLE) {
			// free command buffer
			if (vulkan_commandBuffer)
			{
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, 1, &vulkan_commandBuffer);
				vulkan_commandBuffer = nullptr;
			}

			if (vulkan_commandBuffer1)
			{
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, 1, &vulkan_commandBuffer1);
				vulkan_commandBuffer1 = nullptr;
			}

			//if ((vulkan_PresentQueueCmdBuffers.size() > 0) && (vulkan_PresentQueueCmdBuffers[0] != VK_NULL_HANDLE)) {
			//if (vulkan_PresentQueueCmdBuffers.size() > 0) {
			if (vulkan_PresentQueueCmdBuffers_count > 0) {
				//vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, static_cast<uint32_t>(vulkan_PresentQueueCmdBuffers.size()), &vulkan_PresentQueueCmdBuffers[0]);
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, vulkan_PresentQueueCmdBuffers_count, &vulkan_PresentQueueCmdBuffers[0]);
				//vulkan_PresentQueueCmdBuffers.clear();
				vulkan_PresentQueueCmdBuffers_count = 0;
			}

			if (vulkan_draw_buffer_0[0])
			{
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, VULKAN_SWAP_IMAGE_COUNT, &vulkan_draw_buffer_0[0]);
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, VULKAN_SWAP_IMAGE_COUNT, &vulkan_draw_buffer_256[0]);
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, VULKAN_SWAP_IMAGE_COUNT, &vulkan_draw_buffer_512[0]);
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, VULKAN_SWAP_IMAGE_COUNT, &vulkan_draw_buffer_1024[0]);
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, VULKAN_SWAP_IMAGE_COUNT, &vulkan_draw_buffer_2048[0]);
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, VULKAN_SWAP_IMAGE_COUNT, &vulkan_draw_buffer_4096[0]);

				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, 1, &vulkan_none_buffer_0);
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, 1, &vulkan_none_buffer_256);
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, 1, &vulkan_none_buffer_512);
				vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, 1, &vulkan_none_buffer_1024);

				vulkan_draw_buffer_0[0] = nullptr;
			}

			//if (vulkan_PresentQueueDrawOnlyBuffers_count > 0)
			//{
			//	vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, vulkan_PresentQueueDrawOnlyBuffers_count, &vulkan_PresentQueueDrawOnlyBuffers[0]);
			//	//vulkan_PresentQueueDrawOnlyBuffers.clear();
			//	vulkan_PresentQueueDrawOnlyBuffers_count = 0;
			//}

			// should only create this once and destroy once probably
			//vulkanapi::vkDestroyCommandPool(vulkan_device, vulkan_commandPool, nullptr);
			//vulkan_commandPool = VK_NULL_HANDLE;
		}

	}
}


void vulkan_destroy()
{
	// clear presentation queue stuff
	vulkan_clear();

	if (vulkan_device != VK_NULL_HANDLE)
	{
		vulkanapi::vkDeviceWaitIdle(vulkan_device);

		if (vulkan_mappedmemory)
		{
			vulkanapi::vkUnmapMemory(vulkan_device, vulkan_memory);
			vulkan_mappedmemory = nullptr;
		}

		vulkanapi::vkDestroyFence(vulkan_device, vulkan_fence, nullptr);

		// validation layer: vulkan api says have to create pool with create/free descriptor set flag
		//vulkanapi::vkFreeDescriptorSets(vulkan_device, vulkan_descriptorPool, 1, &vulkan_descriptorSet);
		vulkanapi::vkDestroyDescriptorPool(vulkan_device, vulkan_descriptorPool, nullptr);

		vulkanapi::vkDestroyPipeline(vulkan_device, vulkan_pipeline, nullptr);
		vulkanapi::vkDestroyPipelineLayout(vulkan_device, vulkan_pipelineLayout, nullptr);

		vulkanapi::vkDestroyDescriptorSetLayout(vulkan_device, vulkan_descriptorSetLayout, nullptr);

		vulkanapi::vkDestroyShaderModule(vulkan_device, vulkan_shader_module, nullptr);

		vulkanapi::vkDestroyImage(vulkan_device, vulkan_image, nullptr);

		//for (int i = 0; i < vulkan_buffers.size(); i++)
		for (int i = 0; i < vulkan_buffers_count; i++)
		{
			vulkanapi::vkDestroyBuffer(vulkan_device, vulkan_buffers[i], nullptr);
		}
		vulkanapi::vkFreeMemory(vulkan_device, vulkan_memory, nullptr);
		vulkanapi::vkFreeMemory(vulkan_device, vulkan_memory_local, nullptr);
		vulkanapi::vkFreeMemory(vulkan_device, vulkan_memory_image, nullptr);

		//vulkanapi::vkDestroyBuffer(vulkan_device, vulkan_scratch_buffer, nullptr);
		//vulkanapi::vkDestroyBuffer(vulkan_device, vulkan_inout_buffer, nullptr);
		//vulkanapi::vkDestroyBuffer(vulkan_device, vulkan_out_buffer, nullptr);
		//vulkanapi::vkDestroyBuffer(vulkan_device, vulkan_incomm_buffer, nullptr);
		//vulkanapi::vkDestroyBuffer(vulkan_device, vulkan_indata_buffer, nullptr);
		//vulkanapi::vkDestroyBuffer(vulkan_device, vulkan_work_buffer, nullptr);
		//vulkanapi::vkDestroyBuffer(vulkan_device, vulkan_staging_buffer, nullptr);


		if (vulkan_image_available_semaphore != VK_NULL_HANDLE) {
			vulkanapi::vkDestroySemaphore(vulkan_device, vulkan_image_available_semaphore, nullptr);
			vulkan_image_available_semaphore = VK_NULL_HANDLE;
		}
		if (vulkan_rendering_finished_semaphore != VK_NULL_HANDLE) {
			vulkanapi::vkDestroySemaphore(vulkan_device, vulkan_rendering_finished_semaphore, nullptr);
			vulkan_rendering_finished_semaphore = VK_NULL_HANDLE;
		}
		if (vulkan_swapchain != VK_NULL_HANDLE) {
			//vulkanapi::vkDestroySwapchainKHR(vulkan_device, vulkan_swapchain, nullptr);
			vulkanapi::vkDestroySwapchainKHR(vulkan_device, vulkan_swapchain, vulkan_pAllocator);
			vulkan_swapchain = VK_NULL_HANDLE;

			// deallocate swap chain images
			//swap_chain_images.clear();
			swap_chain_images_count = 0;
		}

		if (vulkan_commandPool != VK_NULL_HANDLE) {
			// free command buffer
			//vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, 1, &vulkan_commandBuffer);
			//vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, 1, &vulkan_commandBuffer1);

			//if ((vulkan_PresentQueueCmdBuffers.size() > 0) && (vulkan_PresentQueueCmdBuffers[0] != VK_NULL_HANDLE)) {
			//if (vulkan_PresentQueueCmdBuffers.size() > 0) {
			//	vulkanapi::vkFreeCommandBuffers(vulkan_device, vulkan_commandPool, static_cast<uint32_t>(vulkan_PresentQueueCmdBuffers.size()), &vulkan_PresentQueueCmdBuffers[0]);
			//	vulkan_PresentQueueCmdBuffers.clear();
			//}

			vulkanapi::vkDestroyCommandPool(vulkan_device, vulkan_commandPool, nullptr);
			vulkan_commandPool = VK_NULL_HANDLE;
		}

		// do this last
		vulkanapi::vkDestroyDevice(vulkan_device, nullptr);
		vulkan_device = VK_NULL_HANDLE;
	}


	if (vulkan_instance != VK_NULL_HANDLE) {
		if (enableValidationLayers) {
			vulkanapi::vkDestroyDebugUtilsMessengerEXT(vulkan_instance, debugMessenger, nullptr);
		}

		if (vulkan_presentation_surface != VK_NULL_HANDLE) {
			//vulkanapi::vkDestroySurfaceKHR(vulkan_instance, vulkan_presentation_surface, nullptr);
			vulkanapi::vkDestroySurfaceKHR(vulkan_instance, vulkan_presentation_surface, vulkan_pAllocator);

			vulkan_presentation_surface = VK_NULL_HANDLE;
		}

		//vulkanapi::vkDestroyInstance(vulkan_instance, nullptr);
		vulkanapi::vkDestroyInstance(vulkan_instance, vulkan_pAllocator);

		vulkan_instance = VK_NULL_HANDLE;
	}

	// delete vectors
	//std::vector<VkBuffer> vulkan_buffers(0);
	//std::vector<VkCommandBuffer> vulkan_PresentQueueCmdBuffers(0);
	//vulkan_buffers.clear();
	//vulkan_PresentQueueCmdBuffers.clear();
	vulkan_buffers_count = 0;
	vulkan_PresentQueueCmdBuffers_count = 0;

	//vulkan_surface_formats.clear();
	//vulkan_present_modes.clear();
	//swap_chain_images.clear();
	vulkan_surface_formats_count = 0;
	vulkan_present_modes_count = 0;
	swap_chain_images_count = 0;

	if (VulkanLibrary) {
#if defined(VK_USE_PLATFORM_WIN32_KHR)
		FreeLibrary(VulkanLibrary);
#elif defined(VK_USE_PLATFORM_XCB_KHR) || defined(VK_USE_PLATFORM_XLIB_KHR)
		dlclose(VulkanLibrary);
#endif
	}

	// vulkan is now unloaded
	vulkan_loaded = false;

	// vulkan commands have been cleared
	vulkan_commands_recorded = VULKAN_COMMANDS_RECORDED_NONE;

}


// it is possible for vulkan to request the swap chain be recreated, like if window size changes etc
// swap chain is used to show the output in the main program window
bool vulkan_create_swap_chain()
{
	cout << "\nVULKAN: INFO: setting up swap chain\n";


	// clear previous swap chain data
	//vulkan_clear();
	//vulkan_surface_formats.clear();
	//vulkan_present_modes.clear();
	//swap_chain_images.clear();
	vulkan_surface_formats_count = 0;
	vulkan_present_modes_count = 0;
	swap_chain_images_count = 0;

	// get surface capabilities
	VkSurfaceCapabilitiesKHR surface_capabilities;
	if (vulkanapi::vkGetPhysicalDeviceSurfaceCapabilitiesKHR(vulkan_physicalDevice, vulkan_presentation_surface, &surface_capabilities) != VK_SUCCESS) {
		std::cout << "ERROR: VULKAN: Could not check presentation surface capabilities!" << std::endl;
		return false;
	}


	// get supported surface formats
	uint32_t formats_count;
	if ((vulkanapi::vkGetPhysicalDeviceSurfaceFormatsKHR(vulkan_physicalDevice, vulkan_presentation_surface, &formats_count, nullptr) != VK_SUCCESS) ||
		(formats_count == 0)) {
		std::cout << "ERROR: VULKAN: occurred during presentation surface formats enumeration!" << std::endl;
		return false;
	}

	//std::vector<VkSurfaceFormatKHR> vulkan_surface_formats(formats_count);
	//vulkan_surface_formats.resize(formats_count);
	vulkan_surface_formats_count = formats_count;
	if (vulkanapi::vkGetPhysicalDeviceSurfaceFormatsKHR(vulkan_physicalDevice, vulkan_presentation_surface, &formats_count, &vulkan_surface_formats[0]) != VK_SUCCESS) {
		std::cout << "ERROR: VULKAN: occurred during presentation surface formats enumeration!" << std::endl;
		return false;
	}


	// get supported present modes
	uint32_t present_modes_count;
	if ((vulkanapi::vkGetPhysicalDeviceSurfacePresentModesKHR(vulkan_physicalDevice, vulkan_presentation_surface, &present_modes_count, nullptr) != VK_SUCCESS) ||
		(present_modes_count == 0)) {
		std::cout << "Error occurred during presentation surface present modes enumeration!" << std::endl;
		return false;
	}

	//std::vector<VkPresentModeKHR> vulkan_present_modes(present_modes_count);
	//vulkan_present_modes.resize(present_modes_count);
	vulkan_present_modes_count = present_modes_count;
	if (vulkanapi::vkGetPhysicalDeviceSurfacePresentModesKHR(vulkan_physicalDevice, vulkan_presentation_surface, &present_modes_count, &vulkan_present_modes[0]) != VK_SUCCESS) {
		std::cout << "Error occurred during presentation surface present modes enumeration!" << std::endl;
		return false;
	}



	// Set of images defined in a swap chain may not always be available for application to render to:
	// One may be displayed and one may wait in a queue to be presented
	// If application wants to use more images at the same time it must ask for more images
	vulkan_image_count = surface_capabilities.minImageCount + 1;
	if ((surface_capabilities.maxImageCount > 0) &&
		(vulkan_image_count > surface_capabilities.maxImageCount)) {
		vulkan_image_count = surface_capabilities.maxImageCount;
	}

	
	std::cout << "\nVULKAN: number of swapchain images: " << dec << vulkan_image_count << " out of a max of: " << surface_capabilities.maxImageCount << " and a MIN of: " << surface_capabilities.minImageCount;

	// select format for swap chain images
	// If the list contains only one entry with undefined format
	// it means that there are no preferred surface formats and any can be chosen
	VkSurfaceFormatKHR surface_format;
	surface_format.format = VK_FORMAT_UNDEFINED;
	//if ((vulkan_surface_formats.size() == 1) &&
	if ((vulkan_surface_formats_count == 1) &&
		(vulkan_surface_formats[0].format == VK_FORMAT_UNDEFINED)) {
		surface_format = { VK_FORMAT_R8G8B8A8_UNORM, VK_COLORSPACE_SRGB_NONLINEAR_KHR };
	}
	else
	{

		// Check if list contains most widely used R8 G8 B8 A8 format
		// with nonlinear color space
		//for (/* VkSurfaceFormatKHR& */ surface_format : surface_formats) {
		//for (int i = 0; i < vulkan_surface_formats.size(); i++)
		for (int i = 0; i < vulkan_surface_formats_count; i++)
		{
			if (vulkan_surface_formats[i].format == VK_FORMAT_R8G8B8A8_UNORM)
			{
				//return surface_format;
				surface_format = vulkan_surface_formats[i];
				break;
			}
		}
	}

	if (surface_format.format != VK_FORMAT_R8G8B8A8_UNORM)
	{
		std::cout << "\nERROR: VULKAN: unable to choose desired swap chain surface format.";

		surface_format = vulkan_surface_formats[0];
	}



	// Return the first format from the list
	//return surface_formats[0];


	// select size of swap chain images
	// Special value of surface extent is width == height == -1
	// If this is so we define the size by ourselves but it must fit within defined confines
	VkExtent2D swap_chain_extent;
	if (surface_capabilities.currentExtent.width == -1) {

		// TODO: this should probably be the size of the program window - not constant
		//swap_chain_extent = { 640, 480 };
		swap_chain_extent = { vulkan_screen_size_x, vulkan_screen_size_y };

		if (swap_chain_extent.width < surface_capabilities.minImageExtent.width) {
			swap_chain_extent.width = surface_capabilities.minImageExtent.width;
		}
		if (swap_chain_extent.height < surface_capabilities.minImageExtent.height) {
			swap_chain_extent.height = surface_capabilities.minImageExtent.height;
		}
		if (swap_chain_extent.width > surface_capabilities.maxImageExtent.width) {
			swap_chain_extent.width = surface_capabilities.maxImageExtent.width;
		}
		if (swap_chain_extent.height > surface_capabilities.maxImageExtent.height) {
			swap_chain_extent.height = surface_capabilities.maxImageExtent.height;
		}
		//return swap_chain_extent;
	}
	else
	{
		swap_chain_extent = surface_capabilities.currentExtent;
	}

	// Most of the cases we define size of the swap_chain images equal to current window's size
	//return surface_capabilities.currentExtent;


	VkCompositeAlphaFlagBitsKHR composite_alpha;
	if (surface_capabilities.supportedCompositeAlpha & VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR) {
		composite_alpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
	}
	else
	{
		std::cout << "ERROR: VULKAN: VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR composite is not supported by the swap chain!" << std::endl
			<< "Supported swap chain's image usages include:" << std::endl
			<< (surface_capabilities.supportedCompositeAlpha & VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR ? "    VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedCompositeAlpha & VK_COMPOSITE_ALPHA_PRE_MULTIPLIED_BIT_KHR ? "    VK_COMPOSITE_ALPHA_PRE_MULTIPLIED_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedCompositeAlpha & VK_COMPOSITE_ALPHA_POST_MULTIPLIED_BIT_KHR ? "    VK_COMPOSITE_ALPHA_POST_MULTIPLIED_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedCompositeAlpha & VK_COMPOSITE_ALPHA_INHERIT_BIT_KHR ? "    VK_COMPOSITE_ALPHA_INHERIT_BIT_KHR\n" : "")
			<< std::endl;
	}


	// Color attachment flag must always be supported
	// We can define other usage flags but we always need to check if they are supported
	VkImageUsageFlags surface_flags = static_cast<VkImageUsageFlags>(-1);
	if (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_TRANSFER_DST_BIT) {
		surface_flags = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_DST_BIT;
	}
	else
	{
		std::cout << "ERROR: VULKAN: VK_IMAGE_USAGE_TRANSFER_DST image usage is not supported by the swap chain!" << std::endl
			<< "Supported swap chain's image usages include:" << std::endl
			<< (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_TRANSFER_SRC_BIT ? "    VK_IMAGE_USAGE_TRANSFER_SRC\n" : "")
			<< (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_TRANSFER_DST_BIT ? "    VK_IMAGE_USAGE_TRANSFER_DST\n" : "")
			<< (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_SAMPLED_BIT ? "    VK_IMAGE_USAGE_SAMPLED\n" : "")
			<< (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_STORAGE_BIT ? "    VK_IMAGE_USAGE_STORAGE\n" : "")
			<< (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT ? "    VK_IMAGE_USAGE_COLOR_ATTACHMENT\n" : "")
			<< (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT ? "    VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT\n" : "")
			<< (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT_BIT ? "    VK_IMAGE_USAGE_TRANSIENT_ATTACHMENT\n" : "")
			<< (surface_capabilities.supportedUsageFlags & VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT ? "    VK_IMAGE_USAGE_INPUT_ATTACHMENT" : "")
			<< std::endl;
	}

	//return static_cast<VkImageUsageFlags>(-1);



	// Sometimes images must be transformed before they are presented (i.e. due to device's orienation
	// being other than default orientation)
	// If the specified transform is other than current transform, presentation engine will transform image
	// during presentation operation; this operation may hit performance on some platforms
	// Here we don't want any transformations to occur so if the identity transform is supported use it
	// otherwise just use the same transform as current transform
	VkSurfaceTransformFlagBitsKHR surface_transform;
	//if (surface_capabilities.supportedTransforms & desired_surface_transform) {
	if (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR) {
				//if (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR) {
			//if (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_180_BIT_KHR) {
			//return VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR;

		surface_transform = VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR;

		//surface_transform = VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_180_BIT_KHR;
		//surface_transform = VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR;
	}
	else {
		//return surface_capabilities.currentTransform;
		surface_transform = surface_capabilities.currentTransform;

		std::cout << "ERROR: VULKAN: VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR transform is not supported by the swap chain!" << std::endl
			<< "Supported swap chain's image transforms include:" << std::endl
			<< (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR ? "    VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR ? "    VK_SURFACE_TRANSFORM_ROTATE_90_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR ? "    VK_SURFACE_TRANSFORM_ROTATE_180_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR ? "    VK_SURFACE_TRANSFORM_ROTATE_270_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_BIT_KHR ? "    VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_90_BIT_KHR ? "    VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_90_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_180_BIT_KHR ? "    VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_180_BIT_KHR\n" : "")
			<< (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_270_BIT_KHR ? "    VK_SURFACE_TRANSFORM_HORIZONTAL_MIRROR_ROTATE_270_BIT_KHR" : "")
			<< (surface_capabilities.supportedTransforms & VK_SURFACE_TRANSFORM_INHERIT_BIT_KHR ? "    VK_SURFACE_TRANSFORM_INHERIT_BIT_KHR" : "")
			<< std::endl;
	}



	// TODO - CONTINUE FROM HERE
	// FIFO present mode is always available
	// MAILBOX is the lowest latency V-Sync enabled mode (something like triple-buffering) so use it if available
	VkPresentModeKHR present_mode = static_cast<VkPresentModeKHR>(-1);
	//for (VkPresentModeKHR& present_mode : present_modes) {
	//for (int i = 0; i < vulkan_present_modes.size(); i++)
	for (int i = 0; i < vulkan_present_modes_count; i++)
	{
		if (vulkan_present_modes[i] == VK_PRESENT_MODE_FIFO_KHR)
		{
			present_mode = vulkan_present_modes[i];
			//return present_mode;
		}


		//if (present_modes[i] == VK_PRESENT_MODE_FIFO_KHR)
		if (vulkan_present_modes[i] == VK_PRESENT_MODE_MAILBOX_KHR)
		{
			present_mode = vulkan_present_modes[i];

			// prefer mailbox present mode over fifo present mode
			break;
			//return present_mode;
		}
	}

	// testing //
	//present_mode = VK_PRESENT_MODE_FIFO_KHR;

	std::cout << "\npresent_mode= " << present_mode;

	if (present_mode == static_cast<VkPresentModeKHR>(-1))
	{
		std::cout << "ERROR: VULKAN: FIFO present mode is not supported by the swap chain!" << std::endl;
		//return static_cast<VkPresentModeKHR>(-1);
		return false;
	}



	// create swap chain
	uint32_t                      desired_number_of_images = vulkan_image_count;	// GetSwapChainNumImages(surface_capabilities);
	VkSurfaceFormatKHR            desired_format = surface_format;	// GetSwapChainFormat(surface_formats);
	VkExtent2D                    desired_extent = swap_chain_extent;	// GetSwapChainExtent(surface_capabilities);
	VkImageUsageFlags             desired_usage = surface_flags;	// GetSwapChainUsageFlags(surface_capabilities);
	VkSurfaceTransformFlagBitsKHR	desired_transform = surface_transform;	// GetSwapChainTransform(surface_capabilities);
	VkPresentModeKHR              desired_present_mode = present_mode;	// GetSwapChainPresentMode(present_modes);
	VkSwapchainKHR                old_swap_chain = vulkan_swapchain;	// Vulkan.SwapChain;

	//if (static_cast<int>(desired_usage) == -1) {
	//	return false;
	//}
	//if (static_cast<int>(desired_present_mode) == -1) {
	//	return false;
	//}

	VkSwapchainCreateInfoKHR swap_chain_create_info = {
	  VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR,  // VkStructureType                sType
	  nullptr,                                      // const void                    *pNext
	  0,                                            // VkSwapchainCreateFlagsKHR      flags
	  vulkan_presentation_surface,                   // VkSurfaceKHR                   surface
	  desired_number_of_images,                     // uint32_t                       minImageCount
	  desired_format.format,                        // VkFormat                       imageFormat
	  desired_format.colorSpace,                    // VkColorSpaceKHR                imageColorSpace
	  desired_extent,                               // VkExtent2D                     imageExtent
	  1,                                            // uint32_t                       imageArrayLayers
	  desired_usage,                                // VkImageUsageFlags              imageUsage
	  VK_SHARING_MODE_EXCLUSIVE,                    // VkSharingMode                  imageSharingMode
	  0,                                            // uint32_t                       queueFamilyIndexCount
	  nullptr,                                      // const uint32_t                *pQueueFamilyIndices
	  desired_transform,                            // VkSurfaceTransformFlagBitsKHR  preTransform
	  composite_alpha,								// VkCompositeAlphaFlagBitsKHR    compositeAlpha
	  desired_present_mode,                         // VkPresentModeKHR               presentMode
	  VK_TRUE,                                      // VkBool32                       clipped
	  old_swap_chain                                // VkSwapchainKHR                 oldSwapchain
	};

	//VkSwapchainKHR vulkan_swapchain;


	std::cout << "\nVULKAN: INFO: creating swapchain.\n";

	VkResult create_swap_result;
	for (int i = 0; i < 16; i++)
	{
		//create_swap_result = vulkanapi::vkCreateSwapchainKHR(vulkan_device, &swap_chain_create_info, nullptr, &vulkan_swapchain);
		create_swap_result = vulkanapi::vkCreateSwapchainKHR(vulkan_device, &swap_chain_create_info, vulkan_pAllocator, &vulkan_swapchain);

		if (create_swap_result == VK_SUCCESS) break;
	}

	if (create_swap_result != VK_SUCCESS) {
		std::cout << "ERROR: VULKAN: Could not create swap chain!" << " Error#" << dec << create_swap_result << ": " << VkResultToCString(create_swap_result);
		return false;
	}
	if (old_swap_chain != VK_NULL_HANDLE) {
		vulkanapi::vkDestroySwapchainKHR(vulkan_device, old_swap_chain, nullptr);
	}

	std::cout << "\nVULKAN: INFO: getting swapchain images.\n";

	// get the images created in the swap chain
	//std::vector<VkImage> swap_chain_images(vulkan_image_count);
	//swap_chain_images.resize(vulkan_image_count);
	swap_chain_images_count = vulkan_image_count;
	if (vulkanapi::vkGetSwapchainImagesKHR(vulkan_device, vulkan_swapchain, &vulkan_image_count, &swap_chain_images[0]) != VK_SUCCESS) {
		std::cout << "ERROR: VULKAN: Could not get swap chain images!" << std::endl;
		return false;
	}

	return true;
}


bool vulkan_is_compute_draw_recorded()
{
	return (vulkan_commands_recorded == VULKAN_COMMANDS_RECORDED_COMPUTE_DRAW);
}

bool vulkan_record_default_commands()
{
	// creating command buffers to execute and draw to each image //

	std::cout << "\nVULKAN: INFO: recording default commands.\n";

	// no commands recorded yet
	vulkan_commands_recorded = VULKAN_COMMANDS_RECORDED_NONE;

	// clear previous swap chain data
	vulkan_clear();


	// *testing* record the new commands
	vulkan_prerecord_command_buffers( vulkan_num_gpu_x_workgroups );


	// create/recreate swap chain
	//if(!vulkan_create_swap_chain())
	//{
	//	std::cout << "\nVULKAN: RECORD COMPUTE+DRAW: ERROR: problem creating SWAP CHAIN.";
	//	return nullptr;
	//}


	// get the buffer to use for staging, which is going to be the last one
	//VkBuffer vulkan_staging_buffer = vulkan_buffers[vulkan_buffers.size() - 1];
	VkBuffer vulkan_staging_buffer = vulkan_buffers[vulkan_buffers_count - 1];

	VkImageSubresourceRange image_subresource_range = { VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 };

	VkClearColorValue clear_color = {
		{ 1.0f, 0.8f, 0.4f, 0.0f }
	};


	VkImageBlit image_blit_region = {
		{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1 },
		{ { 0, 0, 0 }, { 640, 480, 1 } },
		{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1 },
		//{ { 0, 0, 0 }, { 640, 480, 1 } }
		{ { 0, 0, 0 }, { (int32_t)vulkan_screen_size_x, (int32_t)vulkan_screen_size_y, 1 } }
	};


	VkBufferImageCopy region{};
	region.bufferOffset = 0;
	region.bufferRowLength = 0;
	region.bufferImageHeight = 0;

	region.imageSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
	region.imageSubresource.mipLevel = 0;
	region.imageSubresource.baseArrayLayer = 0;
	region.imageSubresource.layerCount = 1;

	region.imageOffset = { 0, 0, 0 };
	region.imageExtent = {
		640,	// width
		480,	// height
		1
	};



	VkBufferMemoryBarrier barrier_wait_until_buffer_written = {
		VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER,
		nullptr,
		VK_ACCESS_SHADER_WRITE_BIT,
		VK_ACCESS_TRANSFER_READ_BIT,
		vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
		vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
		vulkan_staging_buffer,
		0,
		640 * 480 * 4
	};


	VkBufferMemoryBarrier barrier_wait_until_buffer_read = {
		VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER,
		nullptr,
		VK_ACCESS_TRANSFER_READ_BIT,
		VK_ACCESS_SHADER_READ_BIT | VK_ACCESS_SHADER_WRITE_BIT,
		vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
		vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
		vulkan_staging_buffer,
		0,
		640 * 480 * 4
	};

	VkMemoryBarrier barrier_wait_until_memory_written = {
		VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER,
		nullptr,
		VK_ACCESS_SHADER_WRITE_BIT,
		VK_ACCESS_TRANSFER_READ_BIT
	};

	VkCommandBufferAllocateInfo commandBufferAllocateInfo = {
	  VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
	  0,
	  vulkan_commandPool,
	  VK_COMMAND_BUFFER_LEVEL_PRIMARY,
	  1
	};

	VkCommandBufferBeginInfo commandBufferBeginInfo = {
	  VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
	  0,
	  VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT,
	  0
	};

	std::cout << "\nVULKAN: INFO: creating swap chain command buffers\n";

	// create command buffers for vulkan surfaces
	//vulkan_PresentQueueCmdBuffers.resize(vulkan_image_count);
	vulkan_PresentQueueCmdBuffers_count = vulkan_image_count;
	for (int i = 0; i < vulkan_image_count; i++)
	{

		if (vulkanapi::vkAllocateCommandBuffers(vulkan_device, &commandBufferAllocateInfo, &vulkan_PresentQueueCmdBuffers[i]) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem allocating command buffers for swap chain commands.\n";
			return false;
		}

		vulkanapi::vkBeginCommandBuffer(vulkan_PresentQueueCmdBuffers[i], &commandBufferBeginInfo);

		//vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT, VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, 0, 0, nullptr, 1, &barrier_wait_until_buffer_read, 0, nullptr);

		VkBufferCopy bufcopy =
		{
			0, 0, 1024 * 512 * 2
		};
		//vkCmdCopyBuffer(vulkan_PresentQueueCmdBuffers[i], vulkan_out_buffer, vulkan_inout_buffer, 1, &bufcopy);

#ifdef VULKAN_INLCUDE_HOST_INPUT_STAGE
		VkBufferMemoryBarrier barrier_host_buffer_wait[32];
		for (int idx = 0; idx < vulkan_external_vars_count; idx++)
		{
			VkBufferCopy copyRegion{};
			//copyRegion.srcOffset = 0; // Optional
			//copyRegion.dstOffset = 0; // Optional
			copyRegion.size = vulkan_bufferSizes[idx];

			// destination buffer
			barrier_host_buffer_wait[idx] = {
				VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER,
				nullptr,
				VK_ACCESS_MEMORY_WRITE_BIT,
				VK_ACCESS_MEMORY_READ_BIT,
				vulkan_queueFamilyIndex,
				vulkan_queueFamilyIndex,
				vulkan_buffers[idx],
				0,
				VK_WHOLE_SIZE
			};
		}

		// wait for buffer copies to finish
		vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_HOST_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, vulkan_external_vars_count, &barrier_host_buffer_wait[0], 0, nullptr);
#endif

		// I'd imagine the compute shader run should go here
		vulkanapi::vkCmdBindPipeline(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_BIND_POINT_COMPUTE, vulkan_pipeline);

		vulkanapi::vkCmdBindDescriptorSets(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_BIND_POINT_COMPUTE, vulkan_pipelineLayout, 0, 1, &vulkan_descriptorSet, 0, 0);

		// set number of shaders to run simultaneously here
		// original code
		vulkanapi::vkCmdDispatch(vulkan_PresentQueueCmdBuffers[i], vulkan_num_gpu_x_workgroups, 1, 1);

		// testing
		//vulkanapi::vkCmdDispatch(vulkan_PresentQueueCmdBuffers[i], 8, 1, 1);
		//vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, 0, 0, nullptr, 0, &barrier_wait_until_buffer_written, 0, nullptr);
		//vulkanapi::vkCmdDispatch(vulkan_PresentQueueCmdBuffers[i], 1, 1, 1);

		// wait for all previous commands to finish
		//vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 1, &barrier_wait_until_buffer_written, 0, nullptr);


		//vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, &barrier_wait_until_image_written);
		//vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 1, &barrier_wait_until_buffer_written, 0, nullptr);

		// first 6 buffers buffers copied in at beginning, then copied back out at end
		uint32_t barrier_buffer_wait_count = 0;
		VkBufferMemoryBarrier barrier_buffer_wait[32];
#ifdef VULKAN_INLCUDE_BARRIER_STAGE

		for (int idx = 0; idx < vulkan_buffers_count; idx++)
		{

			// destination buffer
			barrier_buffer_wait[idx] = {
				VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER,
				nullptr,
				VK_ACCESS_SHADER_WRITE_BIT,
				VK_ACCESS_TRANSFER_READ_BIT,
				vulkan_queueFamilyIndex,
				vulkan_queueFamilyIndex,
				vulkan_buffers[idx],
				0,
				VK_WHOLE_SIZE
			};
		}
		barrier_buffer_wait_count = vulkan_buffers_count;

#endif


		VkImageMemoryBarrier barrier_from_present_to_clear = {
			VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
			nullptr,                                    // const void                            *pNext
			VK_ACCESS_SHADER_WRITE_BIT,                  // VkAccessFlags                          srcAccessMask
			VK_ACCESS_MEMORY_WRITE_BIT,               // VkAccessFlags                          dstAccessMask
			VK_IMAGE_LAYOUT_UNDEFINED,                  // VkImageLayout                          oldLayout
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,       // VkImageLayout                          newLayout
			vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
			vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
			vulkan_image,                       // VkImage                                image
			image_subresource_range                     // VkImageSubresourceRange                subresourceRange
		};

		// draw to frame buffer
		//vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, &barrier_from_present_to_clear);
		//vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, &barrier_from_present_to_clear);
		//vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, &barrier_from_present_to_clear);

		vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, barrier_buffer_wait_count, &barrier_buffer_wait[0], 1, &barrier_from_present_to_clear);
		//vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, &barrier_from_present_to_clear);

		//vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, &barrier_wait_until_buffer_written, 1, &barrier_from_present_to_clear);




		//vkCmdClearColorImage(vulkan_PresentQueueCmdBuffers[i], swap_chain_images[i], VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, &clear_color, 1, &image_subresource_range);

		//vkCmdCopyBuffer(vulkan_PresentQueueCmdBuffers[i], vulkan_inout_buffer, vulkan_out_buffer, 1, &bufcopy);


		// draw the screen from buffer
		//vulkanapi::vkCmdCopyBufferToImage(vulkan_PresentQueueCmdBuffers[i], vulkan_staging_buffer, swap_chain_images[i], VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &region);
		vulkanapi::vkCmdCopyBufferToImage(vulkan_PresentQueueCmdBuffers[i], vulkan_staging_buffer, vulkan_image, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &region);


		// image barrier for the image that compute shader writes to
		VkImageMemoryBarrier barrier_wait_until_image_written[2] =
		{
		{
			VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
			nullptr,                                    // const void                            *pNext
			VK_ACCESS_MEMORY_WRITE_BIT,                  // VkAccessFlags                          srcAccessMask
			VK_ACCESS_MEMORY_READ_BIT,               // VkAccessFlags                          dstAccessMask
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,                  // VkImageLayout                          oldLayout
			VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,				// VkImageLayout                          newLayout
			vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
			vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
			vulkan_image,                       // VkImage                                image
			image_subresource_range                     // VkImageSubresourceRange                subresourceRange
		},
		{
			VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
			nullptr,                                    // const void                            *pNext
			VK_ACCESS_MEMORY_READ_BIT,                  // VkAccessFlags                          srcAccessMask
			VK_ACCESS_MEMORY_WRITE_BIT,               // VkAccessFlags                          dstAccessMask
			VK_IMAGE_LAYOUT_UNDEFINED,                  // VkImageLayout                          oldLayout
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,				// VkImageLayout                          newLayout
			vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
			vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
			swap_chain_images[i],                       // VkImage                                image
			image_subresource_range                     // VkImageSubresourceRange                subresourceRange
		}
		};


		vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 2, (const VkImageMemoryBarrier*) &barrier_wait_until_image_written);

		// blit image to the screen
		vulkanapi::vkCmdBlitImage(vulkan_PresentQueueCmdBuffers[i], vulkan_image, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, swap_chain_images[i], VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &image_blit_region, VK_FILTER_NEAREST);


		// wait for all previous commands to finish
		//vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT, VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT, 0, 0, nullptr, 0, nullptr, 1, &barrier_wait_until_screen_written);


		VkImageMemoryBarrier barrier_from_clear_to_present =
		{
			VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
			nullptr,                                    // const void                            *pNext
			VK_ACCESS_MEMORY_WRITE_BIT,               // VkAccessFlags                          srcAccessMask
			VK_ACCESS_MEMORY_READ_BIT,                  // VkAccessFlags                          dstAccessMask
			//VK_ACCESS_NONE,                  // VkAccessFlags                          dstAccessMask
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,       // VkImageLayout                          oldLayout
			VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,            // VkImageLayout                          newLayout
			vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
			vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
			swap_chain_images[i],                       // VkImage                                image
			image_subresource_range                     // VkImageSubresourceRange                subresourceRange
		};

		
		//vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, barrier_buffer_wait_count, &barrier_buffer_wait[0], 1, (const VkImageMemoryBarrier*)&barrier_from_clear_to_present);
		vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, (const VkImageMemoryBarrier*)&barrier_from_clear_to_present);





		if (vulkanapi::vkEndCommandBuffer(vulkan_PresentQueueCmdBuffers[i]) != VK_SUCCESS) {
			std::cout << "ERROR: VULKAN: Could not record command buffers!" << std::endl;
			return false;
		}

	}	// end for (int i = 0; i < vulkan_image_count; i++)



	// must be creating the "execute only" command buffer below //


	std::cout << "\nVULKAN: INFO: creating second set of command buffers.\n";



	//VkCommandBuffer commandBuffer;
	if (vulkanapi::vkAllocateCommandBuffers(vulkan_device, &commandBufferAllocateInfo, &vulkan_commandBuffer1) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating command buffers.\n";
		return false;

	}




	if (vulkanapi::vkBeginCommandBuffer(vulkan_commandBuffer1, &commandBufferBeginInfo) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem with begin command buffer.\n";
		return false;

	}




	//vkCmdPipelineBarrier(vulkan_commandBuffer1, VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT, VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, 0, 0, nullptr, 0, nullptr, 0, nullptr);

	vulkanapi::vkCmdBindPipeline(vulkan_commandBuffer1, VK_PIPELINE_BIND_POINT_COMPUTE, vulkan_pipeline);

	vulkanapi::vkCmdBindDescriptorSets(vulkan_commandBuffer1, VK_PIPELINE_BIND_POINT_COMPUTE,
		vulkan_pipelineLayout, 0, 1, &vulkan_descriptorSet, 0, 0);

	// set number of shaders to run simultaneously here
	vulkanapi::vkCmdDispatch(vulkan_commandBuffer1, vulkan_num_gpu_x_workgroups, 1, 1);

	uint32_t barrier_buffer_wait_count = 0;
	VkBufferMemoryBarrier barrier_buffer_wait[32];
#ifdef VULKAN_INCLUDE_DUMMY_DRAW_STAGE

	for (int idx = 0; idx < vulkan_buffers_count; idx++)
	{

		// destination buffer
		barrier_buffer_wait[idx] = {
			VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER,
			nullptr,
			VK_ACCESS_SHADER_WRITE_BIT,
			VK_ACCESS_TRANSFER_READ_BIT,
			vulkan_queueFamilyIndex,
			vulkan_queueFamilyIndex,
			vulkan_buffers[idx],
			0,
			VK_WHOLE_SIZE
		};
	}
	barrier_buffer_wait_count = vulkan_buffers_count;

	VkImageMemoryBarrier barrier_from_present_to_clear = {
		VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
		nullptr,                                    // const void                            *pNext
		VK_ACCESS_SHADER_WRITE_BIT,                  // VkAccessFlags                          srcAccessMask
		VK_ACCESS_MEMORY_WRITE_BIT,               // VkAccessFlags                          dstAccessMask
		VK_IMAGE_LAYOUT_UNDEFINED,                  // VkImageLayout                          oldLayout
		VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,       // VkImageLayout                          newLayout
		vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
		vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
		vulkan_image,                       // VkImage                                image
		image_subresource_range                     // VkImageSubresourceRange                subresourceRange
	};

	// draw to frame buffer

	vulkanapi::vkCmdPipelineBarrier(vulkan_commandBuffer1, VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, barrier_buffer_wait_count, &barrier_buffer_wait[0], 1, &barrier_from_present_to_clear);

	vulkanapi::vkCmdCopyBufferToImage(vulkan_commandBuffer1, vulkan_staging_buffer, vulkan_image, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &region);
#endif



	if (vulkanapi::vkEndCommandBuffer(vulkan_commandBuffer1) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem with end command buffer.\n";
		return false;

	}




	// delete vectors
	//std::vector<VkSurfaceFormatKHR> surface_formats(formats_count);
	//std::vector<VkPresentModeKHR> present_modes(present_modes_count);
	//std::vector<VkImage> swap_chain_images(vulkan_image_count);
	//vulkan_surface_formats.clear();
	//vulkan_present_modes.clear();
	//swap_chain_images.clear();

	// done recording all compute+draw commands
	vulkan_commands_recorded = VULKAN_COMMANDS_RECORDED_COMPUTE_DRAW;

	// ok
	return true;

}


bool vulkan_is_draw_only_recorded()
{
	return (vulkan_commands_recorded == VULKAN_COMMANDS_RECORDED_DRAW_ONLY);
}

bool vulkan_record_draw_only_commands()
{
	// creating command buffers to execute and draw to each image //

	std::cout << "\nVULKAN: INFO: recording draw only commands.\n";

	// no commands recorded yet
	vulkan_commands_recorded = VULKAN_COMMANDS_RECORDED_NONE;

	// clear previous swap chain data
	vulkan_clear();


	// create/recreate swap chain
	//if(!vulkan_create_swap_chain())
	//{
	//	std::cout << "\nVULKAN: RECORD DRAW-ONLY: ERROR: problem creating SWAP CHAIN.";
	//	return nullptr;
	//}


	// get the buffer to use for staging, which is going to be the last one
	//VkBuffer vulkan_staging_buffer = vulkan_buffers[vulkan_buffers.size() - 1];
	//VkBuffer vulkan_staging_buffer = vulkan_buffers[vulkan_buffers_count - 1];
	// this one is using the svram
	VkBuffer vulkan_staging_buffer = vulkan_buffers[3];

	VkImageSubresourceRange image_subresource_range = { VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 };


	VkImageBlit image_blit_region = {
		{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1 },
		{ { 0, 0, 0 }, { (int32_t)vulkan_buffer_size_x, (int32_t)vulkan_buffer_size_y, 1 } },
		{ VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1 },
		//{ { 0, 0, 0 }, { 640, 480, 1 } }
		{ { 0, 0, 0 }, { (int32_t)vulkan_screen_size_x, (int32_t)vulkan_screen_size_y, 1 } }
	};


	VkBufferImageCopy region{};
	region.bufferOffset = 0;
	region.bufferRowLength = 0;
	region.bufferImageHeight = 0;

	region.imageSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
	region.imageSubresource.mipLevel = 0;
	region.imageSubresource.baseArrayLayer = 0;
	region.imageSubresource.layerCount = 1;

	region.imageOffset = { 0, 0, 0 };
	region.imageExtent = {
		vulkan_buffer_size_x,	// width
		vulkan_buffer_size_y,	// height
		1
	};


	VkCommandBufferAllocateInfo commandBufferAllocateInfo = {
	  VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
	  0,
	  vulkan_commandPool,
	  VK_COMMAND_BUFFER_LEVEL_PRIMARY,
	  1
	};

	VkCommandBufferBeginInfo commandBufferBeginInfo = {
	  VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
	  0,
	  VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT,
	  0
	};

	std::cout << "\nVULKAN: INFO: creating swap chain command buffers\n";


	// create command buffers for vulkan surfaces
	vulkan_PresentQueueCmdBuffers_count = vulkan_image_count;
	for (int i = 0; i < vulkan_image_count; i++)
	{
		if (vulkanapi::vkAllocateCommandBuffers(vulkan_device, &commandBufferAllocateInfo, &vulkan_PresentQueueCmdBuffers[i]) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem allocating command buffers for swap chain commands.\n";
			return false;
		}

		vulkanapi::vkBeginCommandBuffer(vulkan_PresentQueueCmdBuffers[i], &commandBufferBeginInfo);

		VkImageMemoryBarrier barrier_from_present_to_clear = {
			VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
			nullptr,                                    // const void                            *pNext
			VK_ACCESS_MEMORY_READ_BIT,                  // VkAccessFlags                          srcAccessMask
			VK_ACCESS_MEMORY_WRITE_BIT,               // VkAccessFlags                          dstAccessMask
			VK_IMAGE_LAYOUT_UNDEFINED,                  // VkImageLayout                          oldLayout
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,       // VkImageLayout                          newLayout
			vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
			vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
			vulkan_image,                       // VkImage                                image
			image_subresource_range                     // VkImageSubresourceRange                subresourceRange
		};


		vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, &barrier_from_present_to_clear);


		// draw the screen from buffer
		vulkanapi::vkCmdCopyBufferToImage(vulkan_PresentQueueCmdBuffers[i], vulkan_staging_buffer, vulkan_image, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &region);


		// image barrier for the image that compute shader writes to
		VkImageMemoryBarrier barrier_wait_until_image_written[2] =
		{
			{
				VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
				nullptr,                                    // const void                            *pNext
				VK_ACCESS_MEMORY_WRITE_BIT,                  // VkAccessFlags                          srcAccessMask
				VK_ACCESS_MEMORY_READ_BIT,               // VkAccessFlags                          dstAccessMask
				VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,                  // VkImageLayout                          oldLayout
				VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,				// VkImageLayout                          newLayout
				vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
				vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
				vulkan_image,                       // VkImage                                image
				image_subresource_range                     // VkImageSubresourceRange                subresourceRange
			},
			{
				VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
				nullptr,                                    // const void                            *pNext
				VK_ACCESS_MEMORY_READ_BIT,                  // VkAccessFlags                          srcAccessMask
				VK_ACCESS_MEMORY_WRITE_BIT,               // VkAccessFlags                          dstAccessMask
				VK_IMAGE_LAYOUT_UNDEFINED,                  // VkImageLayout                          oldLayout
				VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,				// VkImageLayout                          newLayout
				vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
				vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
				swap_chain_images[i],                       // VkImage                                image
				image_subresource_range                     // VkImageSubresourceRange                subresourceRange
			}
		};


		vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 2, (const VkImageMemoryBarrier*)&barrier_wait_until_image_written);

		// blit image to the screen
		vulkanapi::vkCmdBlitImage(vulkan_PresentQueueCmdBuffers[i], vulkan_image, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, swap_chain_images[i], VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &image_blit_region, VK_FILTER_NEAREST);



		VkImageMemoryBarrier barrier_from_clear_to_present =
		{
			VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,     // VkStructureType                        sType
			nullptr,                                    // const void                            *pNext
			VK_ACCESS_MEMORY_WRITE_BIT,               // VkAccessFlags                          srcAccessMask
			VK_ACCESS_MEMORY_READ_BIT,                  // VkAccessFlags                          dstAccessMask
			//VK_ACCESS_NONE,                  // VkAccessFlags                          dstAccessMask
			VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,       // VkImageLayout                          oldLayout
			VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,            // VkImageLayout                          newLayout
			vulkan_queueFamilyIndex,             // uint32_t                               srcQueueFamilyIndex
			vulkan_queueFamilyIndex,             // uint32_t                               dstQueueFamilyIndex
			swap_chain_images[i],                       // VkImage                                image
			image_subresource_range                     // VkImageSubresourceRange                subresourceRange
		};


		vulkanapi::vkCmdPipelineBarrier(vulkan_PresentQueueCmdBuffers[i], VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, (const VkImageMemoryBarrier*)&barrier_from_clear_to_present);


		if (vulkanapi::vkEndCommandBuffer(vulkan_PresentQueueCmdBuffers[i]) != VK_SUCCESS) {
			std::cout << "ERROR: VULKAN: Could not record command buffers!" << std::endl;
			return false;
		}

	}	// end for (int i = 0; i < vulkan_image_count; i++)


	// done recording all draw-only commands
	vulkan_commands_recorded = VULKAN_COMMANDS_RECORDED_DRAW_ONLY;

	// ok
	return true;
}






static VKAPI_ATTR VkBool32 VKAPI_CALL debugCallback(
	VkDebugUtilsMessageSeverityFlagBitsEXT messageSeverity,
	VkDebugUtilsMessageTypeFlagsEXT messageType,
	const VkDebugUtilsMessengerCallbackDataEXT* pCallbackData,
	void* pUserData) {

	std::cerr << "validation layer: " << pCallbackData->pMessage << std::endl;

	return VK_FALSE;
}


bool vulkan_create_context(HWND wHandle, HINSTANCE wInstance)
{
	// make sure that vulkan is not already loaded
	if (vulkan_loaded)
	{
		std::cout << "\n***VULKAN: INFO: VULKAN already loaded ***\n";
		return true;
	}

	// initially clear vars in case of fail
	vulkan_loaded = false;
	vulkan_commands_recorded = VULKAN_COMMANDS_RECORDED_NONE;

	// create an allocator for vulkan functions
	//vulkan_pAllocator = createPAllocator("testing");

	// memory not mapped yet
	vulkan_mappedmemory = nullptr;

	// swap chain not created yet, so no previous swap chain
	vulkan_swapchain = VK_NULL_HANDLE;

	// dynamically load vulkan //

	if (!vulkan_load_library())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load library!\n";
		return false;
	}

	if (!vulkanapi::vulkan_load_exported_entry_points())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load exported entry points!\n";
		return false;

	}

	if (!vulkanapi::vulkan_load_global_level_entry_points())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load global level entry points!\n";
		return false;

	}



	//const std::vector<const char*> validationLayers = {
	const char* validationLayers[] = {
		"VK_LAYER_KHRONOS_validation"
	};



	const char* required_instance_extensions[] = {
	//const std::vector<const char*> required_instance_extensions[] = {
	  VK_KHR_SURFACE_EXTENSION_NAME,
	#if defined(VK_USE_PLATFORM_WIN32_KHR)
	  VK_KHR_WIN32_SURFACE_EXTENSION_NAME
	#elif defined(VK_USE_PLATFORM_XCB_KHR)
	  VK_KHR_XCB_SURFACE_EXTENSION_NAME
	#elif defined(VK_USE_PLATFORM_XLIB_KHR)
	  VK_KHR_XLIB_SURFACE_EXTENSION_NAME
	#endif
	};

	char* instance_extensions[32];

	uint32_t extensionCount = 0;
	vulkanapi::vkEnumerateInstanceExtensionProperties(nullptr, &extensionCount, nullptr);

	std::cout << "\n" << extensionCount << " extensions supported\n";

	std::vector<VkExtensionProperties> extensions(extensionCount);

	vulkanapi::vkEnumerateInstanceExtensionProperties(nullptr, &extensionCount, &extensions[0]);

	std::cout << "\nnumber of extensions: " << dec << extensionCount;
	std::cout << "\navailable extensions:\n";

	//for (const auto& extension : extensions) {
	for (int i = 0; i < extensionCount; i++)
	{
		std::cout << '\t' << extensions[i].extensionName << '\n';

		//instance_extensions.push_back(extension.extensionName);
		instance_extensions[i] = extensions[i].extensionName;
	}


	std::cout << "enabling extensions:\n";
	//for (const auto& extension : instance_extensions) {
	for (int i = 0; i < extensionCount; i++)
	{
		std::cout << '\t' << instance_extensions[i] << '\n';
	}

	const VkApplicationInfo applicationInfo = {
		VK_STRUCTURE_TYPE_APPLICATION_INFO,
		0,
		"VKComputeSample",
		0,
		"",
		0,
		VK_MAKE_VERSION(1, 3, 0)
	};


	if (enableValidationLayers) {
		//createInfo.enabledLayerCount = static_cast<uint32_t>(validationLayers.size());
		//createInfo.ppEnabledLayerNames = validationLayers.data();
	}
	else {
		//createInfo.enabledLayerCount = 0;
	}

	const VkInstanceCreateInfo instanceCreateInfo = {
	  VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
	  0,
	  0,
	  &applicationInfo,

	  // validation layers
#ifdef NDEBUG
	  0,
	  nullptr,
#else
		1,
		validationLayers,
#endif

		// TODO: validate extension availability
		// enable all the extensions, in case we need to use some (like swap chain for double buffering, etc)
		//static_cast<uint32_t>(required_instance_extensions.size()),
		2,
		required_instance_extensions
		//extensionCount,
		//&instance_extensions[0]
	};

	//VkInstance vulkan_instance;
	//if (vulkanapi::vkCreateInstance(&instanceCreateInfo, nullptr, &vulkan_instance) != VK_SUCCESS)
	if (vulkanapi::vkCreateInstance(&instanceCreateInfo, vulkan_pAllocator, &vulkan_instance) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating VULKAN instance.\n";
		return false;
	}


	if (!vulkanapi::vulkan_load_instance_level_entry_points())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load instance level entry points!\n";
		return false;

	}


	VkDebugUtilsMessengerCreateInfoEXT dbgcreateInfo{};
	dbgcreateInfo.sType = VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT;
	dbgcreateInfo.messageSeverity = VK_DEBUG_UTILS_MESSAGE_SEVERITY_VERBOSE_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT;
	dbgcreateInfo.messageType = VK_DEBUG_UTILS_MESSAGE_TYPE_GENERAL_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_TYPE_VALIDATION_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_TYPE_PERFORMANCE_BIT_EXT;
	dbgcreateInfo.pfnUserCallback = debugCallback;
	dbgcreateInfo.pUserData = nullptr; // Optional

	if (vulkanapi::vkCreateDebugUtilsMessengerEXT(vulkan_instance, &dbgcreateInfo, nullptr, &debugMessenger))
	{
		std::cout << "\nVULKAN ERROR: Problem creating debug messenger.\n";
		return false;

	}



	// create presentation surface for WINDOWS OS
	VkWin32SurfaceCreateInfoKHR surface_create_info = {
		VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR,  // VkStructureType                  sType
		nullptr,                                          // const void                      *pNext
		0,                                                // VkWin32SurfaceCreateFlagsKHR     flags
		wInstance,                                  // HINSTANCE                        hinstance
		wHandle                                     // HWND                             hwnd
	};

	//VkSurfaceKHR vulkan_presentation_surface;
	//if (vulkanapi::vkCreateWin32SurfaceKHR(vulkan_instance, &surface_create_info, nullptr, &vulkan_presentation_surface) != VK_SUCCESS)
	if (vulkanapi::vkCreateWin32SurfaceKHR(vulkan_instance, &surface_create_info, vulkan_pAllocator, &vulkan_presentation_surface) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating WIN32 surface.\n";
		return false;
	}




	uint32_t physicalDeviceCount = 0;
	if (vulkanapi::vkEnumeratePhysicalDevices(vulkan_instance, &physicalDeviceCount, 0) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem enumerating VULKAN physical devices.\n";
		return false;
	}

	//VkPhysicalDevice* const physicalDevices = (VkPhysicalDevice*)malloc(
	//	sizeof(VkPhysicalDevice) * physicalDeviceCount);
	//std::vector<VkPhysicalDevice> physicalDevices(physicalDeviceCount);
	VkPhysicalDevice physicalDevices[16];

	if (vulkanapi::vkEnumeratePhysicalDevices(vulkan_instance, &physicalDeviceCount, &physicalDevices[0]) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem enumerating VULKAN physical devices.\n";
		return false;
	}

	std::cout << "\n" << physicalDeviceCount << " physical devices supported\n";


	VkPhysicalDeviceProperties deviceProperties;
	VkPhysicalDeviceFeatures deviceFeatures;

	VkPhysicalDeviceFeatures2 deviceFeatures2;
	VkPhysicalDeviceVulkan11Features deviceFeatures11;
	VkPhysicalDeviceVulkan12Features deviceFeatures12;
	VkPhysicalDeviceVulkan13Features deviceFeatures13;

	VkPhysicalDeviceSubgroupProperties subgroupProperties;
	VkPhysicalDeviceProperties2 physicalDeviceProperties;

	std::cout << "\navailable devices:\n";

	//VkPhysicalDevice vulkan_physicalDevice;

	//for (const auto& device : physicalDevices) {
	for (int i = 0; i < physicalDeviceCount; i++)
	{
		const auto& device = physicalDevices[i];

		vulkanapi::vkGetPhysicalDeviceProperties(device, &deviceProperties);
		vulkanapi::vkGetPhysicalDeviceFeatures(device, &deviceFeatures);

		deviceFeatures13.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES;
		deviceFeatures13.pNext = nullptr;

		deviceFeatures12.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES;
		deviceFeatures12.pNext = &deviceFeatures13;

		deviceFeatures11.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_FEATURES;
		deviceFeatures11.pNext = &deviceFeatures12;

		deviceFeatures2.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2;
		deviceFeatures2.pNext = &deviceFeatures11;

		vulkanapi::vkGetPhysicalDeviceFeatures2(device, &deviceFeatures2);

		subgroupProperties.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_PROPERTIES;
		subgroupProperties.pNext = nullptr;

		physicalDeviceProperties.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2;
		physicalDeviceProperties.pNext = &subgroupProperties;

		vulkanapi::vkGetPhysicalDeviceProperties2(device, &physicalDeviceProperties);


		std::cout << '\t' << deviceProperties.deviceName << '\n';
		std::cout << '\t' << "devicetype:" << deviceProperties.deviceType << '\n';
		std::cout << '\t' << "apiVersion:" << deviceProperties.apiVersion << '\n';
		std::cout << '\t' << "maxComputeSharedMemorySize:" << deviceProperties.limits.maxComputeSharedMemorySize << '\n';
		std::cout << '\t' << "maxComputeWorkGroupCount:" << deviceProperties.limits.maxComputeWorkGroupCount << '\n';
		std::cout << '\t' << "maxComputeWorkGroupInvocations:" << deviceProperties.limits.maxComputeWorkGroupInvocations << '\n';
		std::cout << '\t' << "maxComputeWorkGroupSize:" << deviceProperties.limits.maxComputeWorkGroupSize << '\n';
		std::cout << '\t' << "maxDescriptorSetStorageBuffers:" << deviceProperties.limits.maxDescriptorSetStorageBuffers << '\n';
		std::cout << '\t' << "subgroupSize: " << subgroupProperties.subgroupSize;


		if ((deviceProperties.deviceType == VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU) || (deviceProperties.deviceType == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU))
		{
			std::cout << "\n***Found suitable device***\n";
			vulkan_physicalDevice = device;

			break;
		}
	}


	//std::cout << "\nVULKAN: DEBUG: storageBuffer16BitAccess:" << deviceFeatures11.storageBuffer16BitAccess;
	//std::cout << "\nVULKAN: DEBUG: shaderInt8:" << deviceFeatures12.shaderInt8;
	//std::cout << "\nVULKAN: DEBUG: storageBuffer8BitAccess:" << deviceFeatures12.storageBuffer8BitAccess;
	//std::cout << "\nVULKAN: DEBUG: uniformAndStorageBuffer8BitAccess:" << deviceFeatures12.uniformAndStorageBuffer8BitAccess;


	// record subgroup size
	vulkan_SubgroupSize = subgroupProperties.subgroupSize;


#ifdef VERBOSE_VULKAN_DEVICE_EXTENSION_INFO
	std::vector<const char*> device_extensions;

	uint32_t deviceExtensionCount = 0;
	vulkanapi::vkEnumerateDeviceExtensionProperties(vulkan_physicalDevice, nullptr, &deviceExtensionCount, nullptr);

	std::cout << "\n" << deviceExtensionCount << " device extensions supported\n";

	std::vector<VkExtensionProperties> device_extensions_props(deviceExtensionCount);

	vulkanapi::vkEnumerateDeviceExtensionProperties(vulkan_physicalDevice, nullptr, &deviceExtensionCount, &device_extensions_props[0]);

	std::cout << "available extensions:\n";
	for (const auto& extension : device_extensions_props) {
		std::cout << '\t' << extension.extensionName << '\n';

		device_extensions.push_back(extension.extensionName);
	}

	//std::cout << "enabling extensions:\n";
	//for (const auto& extension : device_extensions) {
	//	std::cout << '\t' << extension << '\n';
	//}
#endif



	uint32_t queueFamilyPropertiesCount = 0;
	vulkanapi::vkGetPhysicalDeviceQueueFamilyProperties(vulkan_physicalDevice, &queueFamilyPropertiesCount, 0);

	//VkQueueFamilyProperties* const queueFamilyProperties = (VkQueueFamilyProperties*)_alloca(
	//	sizeof(VkQueueFamilyProperties) * queueFamilyPropertiesCount);
	//std::vector<VkQueueFamilyProperties> queueFamilyProperties(queueFamilyPropertiesCount);
	VkQueueFamilyProperties queueFamilyProperties[128];

	vulkanapi::vkGetPhysicalDeviceQueueFamilyProperties(vulkan_physicalDevice, &queueFamilyPropertiesCount, &queueFamilyProperties[0]);

	//uint32_t queueFamilyIndex;
	VkBool32 queue_present_support;


#ifdef VERBOSE_VULKAN_QUEUE_INFO
	// output the queue info
	std::cout << "\nVULKAN: INFO: *** AVAILABLE QUEUES ***";
	for (int i = 0; i < queueFamilyPropertiesCount; i++)
	{
		const VkQueueFlags maskedFlags = queueFamilyProperties[i].queueFlags;

		// need to know if presentation is supported
		vulkanapi::vkGetPhysicalDeviceSurfaceSupportKHR(vulkan_physicalDevice, i, vulkan_presentation_surface, &queue_present_support);

		std::cout << "\n\tQUEUE#" << dec << i
			<< "\n\tSupported queue usages include:"
			<< (maskedFlags & VK_QUEUE_GRAPHICS_BIT ? "\n\t\tVK_QUEUE_GRAPHICS_BIT" : "")
			<< (maskedFlags & VK_QUEUE_COMPUTE_BIT ? "\n\t\tVK_QUEUE_COMPUTE_BIT" : "")
			<< (maskedFlags & VK_QUEUE_TRANSFER_BIT ? "\n\t\tVK_QUEUE_TRANSFER_BIT" : "")
			<< (maskedFlags & VK_QUEUE_SPARSE_BINDING_BIT ? "\n\t\tVK_QUEUE_SPARSE_BINDING_BIT" : "")
			<< (maskedFlags & VK_QUEUE_PROTECTED_BIT ? "\n\t\tVK_QUEUE_PROTECTED_BIT" : "")
			<< (queue_present_support ? "\n\t\t*QUEUE PRESENT SUPPORT*" : "")
			<< std::endl;
	}
#endif


	//for (const auto& properties : queueFamilyProperties) {
	for (int i = 0; i < queueFamilyPropertiesCount; i++)
	{
		// mask out the sparse binding bit that we aren't caring about (yet!) and the transfer bit
		//const VkQueueFlags maskedFlags = (~(VK_QUEUE_TRANSFER_BIT | VK_QUEUE_SPARSE_BINDING_BIT) & queueFamilyProperties[i].queueFlags);
		const VkQueueFlags maskedFlags = queueFamilyProperties[i].queueFlags;

		//if (!(VK_QUEUE_GRAPHICS_BIT & maskedFlags) && (VK_QUEUE_COMPUTE_BIT & maskedFlags)) {
		if (
			(VK_QUEUE_COMPUTE_BIT & maskedFlags)

			//&& (VK_QUEUE_TRANSFER_BIT & maskedFlags)

			// can toggle/modify this for testing
			&& (VK_QUEUE_GRAPHICS_BIT & maskedFlags)
			)
		{

			// this is the queueFamilyIndex for the compute queue, the one for the graphics queue (if there is one) could be different
			vulkan_queueFamilyIndex = i;

			// need to know if presentation is supported
			vulkanapi::vkGetPhysicalDeviceSurfaceSupportKHR(vulkan_physicalDevice, i, vulkan_presentation_surface, &queue_present_support);

			std::cout << "\n***Found suitable queue family index for COMPUTE queue***\n";

			// choose this one right away if it also has presentation support
			if (queue_present_support)
			{
				std::cout << "\n***Found suitable queue family index for COMPUTE queue with PRESENTATION support***\n";
				break;
			}
		}
	}

	VkDeviceQueueCreateInfo deviceQueueCreateInfo{};

	float queuePriority = 1.0f;
	deviceQueueCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
	deviceQueueCreateInfo.queueFamilyIndex = vulkan_queueFamilyIndex;
	deviceQueueCreateInfo.queueCount = 1;
	deviceQueueCreateInfo.pQueuePriorities = &queuePriority;
	deviceQueueCreateInfo.flags = 0;
	deviceQueueCreateInfo.pNext = 0;


	const char* required_device_extensions[] = {
	  VK_KHR_SWAPCHAIN_EXTENSION_NAME
	};


	VkDeviceCreateInfo deviceCreateInfo{};

	deviceCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
	deviceCreateInfo.pNext = &deviceFeatures11;

	deviceCreateInfo.pQueueCreateInfos = &deviceQueueCreateInfo;
	deviceCreateInfo.queueCreateInfoCount = 1;

	// enable all the device features
	deviceCreateInfo.pEnabledFeatures = &deviceFeatures;
	//deviceCreateInfo.pEnabledFeatures = 0;


	deviceCreateInfo.enabledExtensionCount = 1;
	deviceCreateInfo.ppEnabledExtensionNames = &required_device_extensions[0];
	//deviceCreateInfo.enabledExtensionCount = static_cast<uint32_t>(device_extensions.size());
	//deviceCreateInfo.ppEnabledExtensionNames = device_extensions.data();
	//deviceCreateInfo.enabledExtensionCount = 0;
	//deviceCreateInfo.ppEnabledExtensionNames = nullptr;

	deviceCreateInfo.enabledLayerCount = 0;
	deviceCreateInfo.flags = 0;


	//VkDevice vulkan_device;
	if (vulkanapi::vkCreateDevice(vulkan_physicalDevice, &deviceCreateInfo, nullptr, &vulkan_device) != VK_SUCCESS) {
		//throw std::runtime_error("failed to create logical device!");
		std::cout << "\nVULKAN ERROR: Problem creating VULKAN device.\n";
		return false;
	}
	else
	{
		std::cout << "\nVULKAN SUCCESS: Successfully created VULKAN device.\n";

	}



	if (!vulkanapi::vulkan_load_device_level_entry_points())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load device level entry points!\n";
		return false;

	}

	// set gpu type integrated or discrete
	vulkan_device_type = deviceProperties.deviceType;


	// show device details
	cout << "\n*** VULKAN SETUP SUCCESSFUL ***";
	cout << "\nDevice Name: " << deviceProperties.deviceName;
	cout << "\nDevice Type: " << ((deviceProperties.deviceType == 1) ? "INTEGRATED-GPU" : "DISCRETE-GPU");
	cout << "\nDevice local subgroup size: " << dec << vulkan_SubgroupSize;
	//cout << "\nDevice number of local subgroups: " << dec << vulkan_NumSubgroups;
	//cout << "\nDevice local workgroup size: " << dec << vulkan_WorkgroupSize;
	//cout << "\nDevice number of total workgroups: " << dec << vulkan_NumWorkgroups;


	return true;
}


void* vulkan_setup_compute(char* shader_file_data, int shader_file_size, long* external_vars_sizes, int external_vars_count, long* internal_vars_sizes, int internal_vars_count)
{
	// create command pool ??
	VkCommandPoolCreateInfo commandPoolCreateInfo = {
	  VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO,
	  0,
	  VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT,
	  vulkan_queueFamilyIndex
	};

	//VkCommandPool vulkan_commandPool;
	if (vulkanapi::vkCreateCommandPool(vulkan_device, &commandPoolCreateInfo, 0, &vulkan_commandPool) != VK_SUCCESS)
	{
		cout << "\nVULKAN ERROR: Problem creating command pool.\n";
		return false;
	}


	VkQueue computeQueue;
	vulkanapi::vkGetDeviceQueue(vulkan_device, vulkan_queueFamilyIndex, 0, &computeQueue);

	// *** device created *** //


	// create semaphores
	VkSemaphoreCreateInfo semaphore_create_info = {
	  VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,      // VkStructureType          sType
	  nullptr,                                      // const void*              pNext
	  0                                             // VkSemaphoreCreateFlags   flags
	};

	//VkSemaphore vulkan_image_available_semaphore;
	//VkSemaphore vulkan_rendering_finished_semaphore;
	if ((vulkanapi::vkCreateSemaphore(vulkan_device, &semaphore_create_info, nullptr, &vulkan_image_available_semaphore) != VK_SUCCESS) ||
		(vulkanapi::vkCreateSemaphore(vulkan_device, &semaphore_create_info, nullptr, &vulkan_rendering_finished_semaphore) != VK_SUCCESS)) {

		std::cout << "ERROR: VULKAN: Could not create semaphores!" << std::endl;
		return nullptr;
	}






	VkPhysicalDeviceMemoryProperties properties;

	vulkanapi::vkGetPhysicalDeviceMemoryProperties(vulkan_physicalDevice, &properties);


	VkImageCreateInfo vulkan_image_info =
	{
		VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO,
		nullptr,
		0,
		VK_IMAGE_TYPE_2D,

		// I want a 24-bit rgb format, but TODO - need to check to make sure format is supported
		//VK_FORMAT_R32G32B32A32_SFLOAT,	//VK_FORMAT_R8G8B8A8_UINT,
		VK_FORMAT_R8G8B8A8_UNORM,

		// size of the image
		{ VULKAN_IMAGE_WIDTH, VULKAN_IMAGE_HEIGHT, 1 },

		1, 1,
		VK_SAMPLE_COUNT_1_BIT,
		//VK_IMAGE_TILING_LINEAR,
		VK_IMAGE_TILING_OPTIMAL,

		// just storing image for use in compute shader for now
		VK_IMAGE_USAGE_STORAGE_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT | VK_IMAGE_USAGE_TRANSFER_DST_BIT,

		// only using compute queue for now
		VK_SHARING_MODE_EXCLUSIVE,

		// this is just going to be on the compute queue for now, TODO might also need to add graphics queue later for displaying onscreen?
		1, &vulkan_queueFamilyIndex,

		// validation layer: vulkan specification says that cannot init image with VK_IMAGE_LAYOUT_GENERAL
		VK_IMAGE_LAYOUT_UNDEFINED
	};


	//VkImage vulkan_image;
	if (vulkanapi::vkCreateImage(vulkan_device, &vulkan_image_info, nullptr, &vulkan_image) != VK_SUCCESS)
	{
		cout << "\nERROR: VULKAN: problem creating image.";
	}

	VkMemoryRequirements image_memReqs;
	vulkanapi::vkGetImageMemoryRequirements(vulkan_device, vulkan_image, &image_memReqs);


#ifdef VERBOSE_VULKAN_MEMORY_INFO
	for (int i = 0; i < properties.memoryTypeCount; i++)
	{
		const VkMemoryPropertyFlags maskedFlags = properties.memoryTypes[i].propertyFlags;

		std::cout << "\n\tMEMORY INDEX#" << dec << i
			<< "\n\tMemory Size:" << dec << properties.memoryHeaps[properties.memoryTypes[i].heapIndex].size
			<< "\n\tSupported memory usages include:"
			<< (maskedFlags & VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT ? "\n\t\tVK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT ? "\n\t\tVK_MEMORY_PROPERTY_HOST_VISIBLE_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_HOST_COHERENT_BIT ? "\n\t\tVK_MEMORY_PROPERTY_HOST_COHERENT_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_HOST_CACHED_BIT ? "\n\t\tVK_MEMORY_PROPERTY_HOST_CACHED_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_LAZILY_ALLOCATED_BIT ? "\n\t\tVK_MEMORY_PROPERTY_LAZILY_ALLOCATED_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_PROTECTED_BIT ? "\n\t\tVK_MEMORY_PROPERTY_PROTECTED_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_DEVICE_COHERENT_BIT_AMD ? "\n\t\tVK_MEMORY_PROPERTY_DEVICE_COHERENT_BIT_AMD" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_DEVICE_UNCACHED_BIT_AMD ? "\n\t\tVK_MEMORY_PROPERTY_DEVICE_UNCACHED_BIT_AMD" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_RDMA_CAPABLE_BIT_NV ? "\n\t\tVK_MEMORY_PROPERTY_RDMA_CAPABLE_BIT_NV" : "")
			<< std::endl;
	}
#endif


	int total_vars_count = 0;
	//std::vector<VkDeviceSize> vulkan_bufferSizes(external_vars_count + internal_vars_count);

	// get size of external memory buffers //

	vulkan_external_vars_count = external_vars_count;

	// add up external memory total size
	VkDeviceSize memorySize = 0;
	for (int i = 0; i < external_vars_count; i++)
	{
		vulkan_bufferSizes[total_vars_count++] = external_vars_sizes[i];
		memorySize += external_vars_sizes[i];
	}


	// get size of local memory buffers //

	// note: if discrete gpu, then this includes external buffers

	VkDeviceSize memorySize_local = 0;

	// add up internal memory total size
	for (int i = 0; i < internal_vars_count; i++)
	{
		vulkan_bufferSizes[total_vars_count++] = internal_vars_sizes[i];
		memorySize_local += internal_vars_sizes[i];
	}

	// get total count of vars
	//internal_vars_count += external_vars_count;


	// show total number of vars
	std::cout << "\nVULKAN: INFO: total number of vars: " << dec << total_vars_count;
	std::cout << "\n\tnumber of external vars: " << dec << external_vars_count;
	std::cout << "\n\tnumber of internal vars: " << dec << internal_vars_count;


	// output memory types availables


	// set memoryTypeIndex to an invalid entry in the properties.memoryTypes array
	uint32_t memoryTypeIndex = VK_MAX_MEMORY_TYPES;
	if (vulkan_device_type == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU)
	{
		for (uint32_t k = 0; k < properties.memoryTypeCount; k++)
		{
			if (
				(VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_COHERENT_BIT & properties.memoryTypes[k].propertyFlags) &&
				//(VK_MEMORY_PROPERTY_HOST_CACHED_BIT & properties.memoryTypes[k].propertyFlags) &&
				(memorySize < properties.memoryHeaps[properties.memoryTypes[k].heapIndex].size)) {
				memoryTypeIndex = k;
				break;
			}
		}
	}
	else
	{
		for (uint32_t k = 0; k < properties.memoryTypeCount; k++)
		{
			if (
				//(VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_COHERENT_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_CACHED_BIT & properties.memoryTypes[k].propertyFlags) &&
				(memorySize < properties.memoryHeaps[properties.memoryTypes[k].heapIndex].size)) {
				memoryTypeIndex = k;
				break;
			}
		}
	}

	if (memoryTypeIndex == VK_MAX_MEMORY_TYPES)
	{
		std::cout << "\nERROR: VULKAN: Problem selecting external memory type.\n";
		return nullptr;
	}


	uint32_t memoryTypeIndex_local = VK_MAX_MEMORY_TYPES;
	for (uint32_t k = 0; k < properties.memoryTypeCount; k++) {
		if (
			(VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT & properties.memoryTypes[k].propertyFlags) &&
			//(VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT & properties.memoryTypes[k].propertyFlags) &&
			//(VK_MEMORY_PROPERTY_HOST_COHERENT_BIT & properties.memoryTypes[k].propertyFlags) &&
			//(VK_MEMORY_PROPERTY_HOST_CACHED_BIT & properties.memoryTypes[k].propertyFlags) &&
			(memorySize_local < properties.memoryHeaps[properties.memoryTypes[k].heapIndex].size)) {
			memoryTypeIndex_local = k;
			break;
		}
	}

	if (memoryTypeIndex_local == VK_MAX_MEMORY_TYPES)
	{
		std::cout << "\nERROR: VULKAN: Problem selecting local memory type.\n";
		return nullptr;
	}



	cout << "\nEXTERNAL memoryTypeIndex=" << dec << memoryTypeIndex << " memoryTypeCount=" << properties.memoryTypeCount << " VK_MAX_MEMORY_TYPES=" << VK_MAX_MEMORY_TYPES;
	cout << "\nLOCAL memoryTypeIndex=" << dec << memoryTypeIndex_local << " memoryTypeCount=" << properties.memoryTypeCount << " VK_MAX_MEMORY_TYPES=" << VK_MAX_MEMORY_TYPES;


	// allocate external memory //


	const VkMemoryAllocateInfo memoryAllocateInfo = {
		  VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
		  0,
		  memorySize,
		  memoryTypeIndex
	};

	//VkDeviceMemory vulkan_memory;
	if (vulkanapi::vkAllocateMemory(vulkan_device, &memoryAllocateInfo, 0, &vulkan_memory) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating memory for device.\n";
		return nullptr;

	}


	// allocate local memory //

	// note: if discrete gpu, then this includes external vars

	const VkMemoryAllocateInfo memoryAllocateInfo_local = {
		  VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
		  0,
		  memorySize_local,
		  memoryTypeIndex_local
	};

	//VkDeviceMemory vulkan_memory;
	if (vulkanapi::vkAllocateMemory(vulkan_device, &memoryAllocateInfo_local, 0, &vulkan_memory_local) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating local memory for device.\n";
		return nullptr;

	}


	// allocate memory for image //


	const VkMemoryAllocateInfo memoryAllocateInfo_image = {
		  VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
		  0,
		  image_memReqs.size,
		  memoryTypeIndex_local
	};

	//VkDeviceMemory vulkan_memory;
	if (vulkanapi::vkAllocateMemory(vulkan_device, &memoryAllocateInfo_image, 0, &vulkan_memory_image) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating local image memory for device.\n";
		return nullptr;

	}

	vulkanapi::vkBindImageMemory(vulkan_device, vulkan_image, vulkan_memory_image, 0);


	int current_offset = 0;

	//vulkan_bufferCreateInfos.resize(total_vars_count);
	//vulkan_buffers.resize(total_vars_count);
	vulkan_buffers_count = total_vars_count;

	//std::vector<VkBufferCreateInfo> vulkan_bufferCreateInfos(total_vars_count);
	//std::vector<VkDescriptorSetLayoutBinding> descriptorSetLayoutBindings(total_vars_count);
	VkBufferCreateInfo vulkan_bufferCreateInfos[32];
	VkDescriptorSetLayoutBinding descriptorSetLayoutBindings[32];
	uint32_t descriptorSetLayoutBindings_count = 0;

	for (uint32_t i = 0; i < external_vars_count; i++)
	{
		//std::vector<VkBufferCreateInfo> vulkan_bufferCreateInfos(0);
		//std::vector<VkBuffer> vulkan_buffers(0);

		vulkan_bufferCreateInfos[i] = {
		  VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
		  0,
		  0,
		  vulkan_bufferSizes[i],
		  VK_BUFFER_USAGE_STORAGE_BUFFER_BIT | VK_BUFFER_USAGE_TRANSFER_SRC_BIT | VK_BUFFER_USAGE_TRANSFER_DST_BIT,
		  VK_SHARING_MODE_EXCLUSIVE,
		  1,
		  &vulkan_queueFamilyIndex
		};

		if (vulkanapi::vkCreateBuffer(vulkan_device, &vulkan_bufferCreateInfos[i], 0, &vulkan_buffers[i]) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem creating input buffer.\n";
			return nullptr;

		}

		//if (vulkanapi::vkBindBufferMemory(vulkan_device, vulkan_buffers[i], vulkan_memory, 0) != VK_SUCCESS)
		if (vulkanapi::vkBindBufferMemory(vulkan_device, vulkan_buffers[i], vulkan_memory, current_offset) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem binding internal vram buffer.\n";
			return nullptr;

		}

		current_offset += vulkan_bufferSizes[i];

		descriptorSetLayoutBindings[descriptorSetLayoutBindings_count] = { descriptorSetLayoutBindings_count, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 1, VK_SHADER_STAGE_COMPUTE_BIT, 0 };
		descriptorSetLayoutBindings_count++;
	}


	current_offset = 0;
	for (uint32_t i = external_vars_count; i < total_vars_count; i++)
	{
		//std::vector<VkBufferCreateInfo> vulkan_bufferCreateInfos(0);
		//std::vector<VkBuffer> vulkan_buffers(0);

		vulkan_bufferCreateInfos[i] = {
		  VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
		  0,
		  0,
		  vulkan_bufferSizes[i],
		  VK_BUFFER_USAGE_STORAGE_BUFFER_BIT | VK_BUFFER_USAGE_TRANSFER_SRC_BIT | VK_BUFFER_USAGE_TRANSFER_DST_BIT,
		  VK_SHARING_MODE_EXCLUSIVE,
		  1,
		  &vulkan_queueFamilyIndex
		};

		if (vulkanapi::vkCreateBuffer(vulkan_device, &vulkan_bufferCreateInfos[i], 0, &vulkan_buffers[i]) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem creating input buffer.\n";
			return nullptr;

		}

		//if (vulkanapi::vkBindBufferMemory(vulkan_device, vulkan_buffers[i], vulkan_memory, 0) != VK_SUCCESS)
		if (vulkanapi::vkBindBufferMemory(vulkan_device, vulkan_buffers[i], vulkan_memory_local, current_offset) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem binding internal vram buffer.\n";
			return nullptr;

		}

		current_offset += vulkan_bufferSizes[i];

		descriptorSetLayoutBindings[descriptorSetLayoutBindings_count] = { descriptorSetLayoutBindings_count, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 1, VK_SHADER_STAGE_COMPUTE_BIT, 0 };
		descriptorSetLayoutBindings_count++;
	}



	int32_t* payload;



	std::cout << "\nVULKAN: INFO: Creating buffers and binding memory.\n";




	std::cout << "\nVULKAN: INFO: loading and creating compute shader.\n";


	// *** load shader here *** //

	VkShaderModuleCreateInfo shaderModuleCreateInfo = {
		  VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
		  0,
		  0,
		(size_t)shader_file_size,
		reinterpret_cast<const uint32_t*>(shader_file_data)
	};


	//VkShaderModule vulkan_shader_module;

	if (vulkanapi::vkCreateShaderModule(vulkan_device, &shaderModuleCreateInfo, 0, &vulkan_shader_module) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating shader module.\n";
		return nullptr;

	}


	std::cout << "\nVULKAN: INFO: creating layouts.\n";




	VkDescriptorSetLayoutCreateInfo descriptorSetLayoutCreateInfo = {
	  VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO,
	  //0, 0, 7, descriptorSetLayoutBindings
	  0, 0, descriptorSetLayoutBindings_count, &descriptorSetLayoutBindings[0]
	};


	//VkDescriptorSetLayout vulkan_descriptorSetLayout;
	if (vulkanapi::vkCreateDescriptorSetLayout(vulkan_device, &descriptorSetLayoutCreateInfo, 0, &vulkan_descriptorSetLayout) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating descriptor set layout.\n";
		return nullptr;

	}


	VkPipelineLayoutCreateInfo pipelineLayoutCreateInfo = {
	  VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO,
	  0,
	  0,
	  1,
	  &vulkan_descriptorSetLayout,
	  0,
	  0
	};

	//VkPipelineLayout vulkan_pipelineLayout;
	if (vulkanapi::vkCreatePipelineLayout(vulkan_device, &pipelineLayoutCreateInfo, 0, &vulkan_pipelineLayout) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating pipeline layout.\n";
		return nullptr;

	}

	VkComputePipelineCreateInfo computePipelineCreateInfo = {
	  VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO,
	  nullptr, 0,
	  {
		VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
		nullptr, 0, VK_SHADER_STAGE_COMPUTE_BIT, vulkan_shader_module, "main", nullptr
	  },
	  vulkan_pipelineLayout, VK_NULL_HANDLE, 0
	};

	//VkPipeline vulkan_pipeline;
	VkResult vkres;
	vkres = vulkanapi::vkCreateComputePipelines(vulkan_device, VK_NULL_HANDLE, 1, &computePipelineCreateInfo, 0, &vulkan_pipeline);
	//if (vkCreateComputePipelines(vulkan_device, 0, 1, &computePipelineCreateInfo, 0, &vulkan_pipeline) != VK_SUCCESS)
	if (vkres != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating compute pipelines.\n";
		std::cout << "\nvkres=" << vkres;
		return nullptr;

	}



	VkDescriptorPoolSize descriptorPoolSize[1] = {
		{
			VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
			descriptorSetLayoutBindings_count	// 7
		}
	};

	//VkDescriptorPoolSize descriptorPoolSize = {
	// VK_DESCRIPTOR_TYPE_STORAGE_IMAGE,
	 // 1
	//};

	VkDescriptorPoolCreateInfo descriptorPoolCreateInfo = {
	  VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO,
	  0,
	  0,
	  1,	// 1,
	  1,	// 1,
	  descriptorPoolSize
	};

	//VkDescriptorPool vulkan_descriptorPool;
	if (vulkanapi::vkCreateDescriptorPool(vulkan_device, &descriptorPoolCreateInfo, 0, &vulkan_descriptorPool) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating descriptor pool.\n";
		return nullptr;

	}

	VkDescriptorSetAllocateInfo descriptorSetAllocateInfo = {
	  VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO,
	  0,
	  vulkan_descriptorPool,
	  1,
	  &vulkan_descriptorSetLayout
	};

	//VkDescriptorSet vulkan_descriptorSet;
	if (vulkanapi::vkAllocateDescriptorSets(vulkan_device, &descriptorSetAllocateInfo, &vulkan_descriptorSet) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating descriptor sets.\n";
		return nullptr;

	}


	//std::vector<VkDescriptorBufferInfo> descriptorBufferInfo(total_vars_count);
	//std::vector<VkWriteDescriptorSet> writeDescriptorSet(total_vars_count);
	VkDescriptorBufferInfo descriptorBufferInfo[32];
	VkWriteDescriptorSet writeDescriptorSet[32];

	uint32_t descriptor_start_idx = 0;

	uint32_t writeDescriptorSet_count = 0;
	for (uint32_t i = descriptor_start_idx; i < total_vars_count; i++)
	{
		descriptorBufferInfo[writeDescriptorSet_count] = {
			vulkan_buffers[i],
			0,
			VK_WHOLE_SIZE
		};

		writeDescriptorSet[writeDescriptorSet_count] = {
			VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
			0,
			vulkan_descriptorSet,
			writeDescriptorSet_count,
			0,
			1,
			VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
			nullptr,
			&descriptorBufferInfo[writeDescriptorSet_count],
			nullptr
		};

		writeDescriptorSet_count++;
	}




	std::cout << "\nVULKAN: INFO: creating descriptor set.\n";




	//vulkanapi::vkUpdateDescriptorSets(vulkan_device, 7, writeDescriptorSet, 0, 0);
	vulkanapi::vkUpdateDescriptorSets(vulkan_device, writeDescriptorSet_count, &writeDescriptorSet[0], 0, 0);








	//VkSubmitInfo submitInfo;
	vulkan_submitInfo = {
	  VK_STRUCTURE_TYPE_SUBMIT_INFO,
	  0,
	  0,
	  0,
	  0,
	  1,
	  &vulkan_commandBuffer1,
	  0,
	  0
	};


	//VkQueue queue;
	vulkanapi::vkGetDeviceQueue(vulkan_device, vulkan_queueFamilyIndex, 0, &vulkan_queue);

	// create fence
	VkFenceCreateInfo fenceInfo{};
	fenceInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
	fenceInfo.flags = VK_FENCE_CREATE_SIGNALED_BIT;

	if (vulkanapi::vkCreateFence(vulkan_device, &fenceInfo, nullptr, &vulkan_fence) != VK_SUCCESS) {
		std::cout << "\nVULKAN: ERROR: failed to create fence!";
		return nullptr;
	}


	std::cout << "\nVULKAN: INFO: mapping memory for compute shader\n";

	//if (vkMapMemory(vulkan_device, vulkan_memory, 0, memorySize - work_buffer_size_bytes - image_memReqs.size, 0, (void**)&payload) != VK_SUCCESS)
	if (vulkanapi::vkMapMemory(vulkan_device, vulkan_memory, 0, memorySize, 0, (void**)&payload) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem mapping memory for device.\n";
		return nullptr;

	}


	// set the mapped memory
	vulkan_mappedmemory = payload;


	std::cout << "\nVULKAN: INFO: creating swap chain.\n";

	// create swap chain
	if (!vulkan_create_swap_chain())
	{
		std::cout << "\nVULKAN: SETUP: ERROR: problem creating SWAP CHAIN.";
		return nullptr;
	}



	// record the default command buffers to use
	//if (!vulkan_record_default_commands())
	//{
	//	std::cout << "\nVULKAN: ERROR: Problem recording command buffers.\n";
	//	return false;
	//}





	// vulkan loaded
	vulkan_loaded = true;


	std::cout << "\n***VULKAN SETUP COMPLETED SUCCESSFULLY***.\n";
	std::cout << "\nvulkan alloc count: " << vulkan_alloc_count;
	std::cout << "\nvulkan alloc size: " << vulkan_alloc_size;

	// return mapped memory for device
	return (void*)payload;

}


// returns a pointer into the memory for device
// shader_file_name - name of the file that contains the computer shader
// inout_buffer_size_bytes - size of ps1/ps2 gpu ram (1MB or 4MB, etc)
// out_buffer_size_bytes - size of ps1/ps2 screen output
// in_comm_buffer_size_bytes - size of input draw commands buffer
// in_data_buffer_size_bytes - size of input data buffer (graphics data)
// work_buffer_size_bytes - size of work buffer for rasterizing primitives (number of input command buffer slots times data needed per slot)
//void* vulkan_setup(HWND wHandle, HINSTANCE wInstance, char* shader_file_data, int shader_file_size, int inout_buffer_size_bytes, int out_buffer_size_bytes, int in_comm_buffer_size_bytes, int in_data_buffer_size_bytes, int work_buffer_size_bytes)
void* vulkan_setup(HWND wHandle, HINSTANCE wInstance, char* shader_file_data, int shader_file_size, long* external_vars_sizes, int external_vars_count, long* internal_vars_sizes, int internal_vars_count)
{
	// make sure that vulkan is not already loaded
	if (vulkan_loaded)
	{
		std::cout << "\n***VULKAN: INFO: VULKAN already loaded ***\n";
		return vulkan_mappedmemory;
	}

	// initially clear vars in case of fail
	vulkan_loaded = false;
	vulkan_commands_recorded = VULKAN_COMMANDS_RECORDED_NONE;

	// create an allocator for vulkan functions
	//vulkan_pAllocator = createPAllocator("testing");

	// memory not mapped yet
	vulkan_mappedmemory = nullptr;

	// swap chain not created yet, so no previous swap chain
	vulkan_swapchain = VK_NULL_HANDLE;

	// dynamically load vulkan //

	if (!vulkan_load_library())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load library!\n";
		return nullptr;
	}

	if (!vulkanapi::vulkan_load_exported_entry_points())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load exported entry points!\n";
		return nullptr;

	}

	if (!vulkanapi::vulkan_load_global_level_entry_points())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load global level entry points!\n";
		return nullptr;

	}



	//const std::vector<const char*> validationLayers = {
	const char* validationLayers[] = {
		"VK_LAYER_KHRONOS_validation"
	};



	const char* required_instance_extensions[] = {
	  VK_KHR_SURFACE_EXTENSION_NAME,
	#if defined(VK_USE_PLATFORM_WIN32_KHR)
	  VK_KHR_WIN32_SURFACE_EXTENSION_NAME
	#elif defined(VK_USE_PLATFORM_XCB_KHR)
	  VK_KHR_XCB_SURFACE_EXTENSION_NAME
	#elif defined(VK_USE_PLATFORM_XLIB_KHR)
	  VK_KHR_XLIB_SURFACE_EXTENSION_NAME
	#endif
	};

	char* instance_extensions[32];

	uint32_t extensionCount = 0;
	vulkanapi::vkEnumerateInstanceExtensionProperties(nullptr, &extensionCount, nullptr);

	std::cout << "\n" << extensionCount << " extensions supported\n";

	std::vector<VkExtensionProperties> extensions(extensionCount);

	vulkanapi::vkEnumerateInstanceExtensionProperties(nullptr, &extensionCount, &extensions[0]);

	std::cout << "\nnumber of extensions: " << dec << extensionCount;
	std::cout << "\navailable extensions:\n";

	//for (const auto& extension : extensions) {
	for(int i = 0; i < extensionCount; i++)
	{
		std::cout << '\t' << extensions[i].extensionName << '\n';

		//instance_extensions.push_back(extension.extensionName);
		instance_extensions[i] = extensions[i].extensionName;
	}


	std::cout << "enabling extensions:\n";
	//for (const auto& extension : instance_extensions) {
	for (int i = 0; i < extensionCount; i++)
	{
		std::cout << '\t' << instance_extensions[i] << '\n';
	}

	const VkApplicationInfo applicationInfo = {
		VK_STRUCTURE_TYPE_APPLICATION_INFO,
		0,
		"VKComputeSample",
		0,
		"",
		0,
		VK_MAKE_VERSION(1, 3, 0)
	};


	if (enableValidationLayers) {
		//createInfo.enabledLayerCount = static_cast<uint32_t>(validationLayers.size());
		//createInfo.ppEnabledLayerNames = validationLayers.data();
	}
	else {
		//createInfo.enabledLayerCount = 0;
	}

	const VkInstanceCreateInfo instanceCreateInfo = {
	  VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
	  0,
	  0,
	  &applicationInfo,

	  // validation layers
#ifdef NDEBUG
	  0,
	  nullptr,
#else
		1,
		validationLayers,
#endif

	  // TODO: validate extension availability
	  // enable all the extensions, in case we need to use some (like swap chain for double buffering, etc)
	  //static_cast<uint32_t>(required_instance_extensions.size()),
	  //required_instance_extensions.data()
	  extensionCount,
	  &instance_extensions[0]
	};

	//VkInstance vulkan_instance;
	//if (vulkanapi::vkCreateInstance(&instanceCreateInfo, nullptr, &vulkan_instance) != VK_SUCCESS)
	if (vulkanapi::vkCreateInstance(&instanceCreateInfo, vulkan_pAllocator, &vulkan_instance) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating VULKAN instance.\n";
		return nullptr;
	}


	if (!vulkanapi::vulkan_load_instance_level_entry_points())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load instance level entry points!\n";
		return nullptr;

	}


	VkDebugUtilsMessengerCreateInfoEXT dbgcreateInfo{};
	dbgcreateInfo.sType = VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT;
	dbgcreateInfo.messageSeverity = VK_DEBUG_UTILS_MESSAGE_SEVERITY_VERBOSE_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT;
	dbgcreateInfo.messageType = VK_DEBUG_UTILS_MESSAGE_TYPE_GENERAL_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_TYPE_VALIDATION_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_TYPE_PERFORMANCE_BIT_EXT;
	dbgcreateInfo.pfnUserCallback = debugCallback;
	dbgcreateInfo.pUserData = nullptr; // Optional

	if (vulkanapi::vkCreateDebugUtilsMessengerEXT(vulkan_instance, &dbgcreateInfo, nullptr, &debugMessenger))
	{
		std::cout << "\nVULKAN ERROR: Problem creating debug messenger.\n";
		return nullptr;

	}



	// create presentation surface for WINDOWS OS
	VkWin32SurfaceCreateInfoKHR surface_create_info = {
		VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR,  // VkStructureType                  sType
		nullptr,                                          // const void                      *pNext
		0,                                                // VkWin32SurfaceCreateFlagsKHR     flags
		wInstance,                                  // HINSTANCE                        hinstance
		wHandle                                     // HWND                             hwnd
	};

	//VkSurfaceKHR vulkan_presentation_surface;
	//if (vulkanapi::vkCreateWin32SurfaceKHR(vulkan_instance, &surface_create_info, nullptr, &vulkan_presentation_surface) != VK_SUCCESS)
	if (vulkanapi::vkCreateWin32SurfaceKHR(vulkan_instance, &surface_create_info, vulkan_pAllocator, &vulkan_presentation_surface) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating WIN32 surface.\n";
		return nullptr;
	}




	uint32_t physicalDeviceCount = 0;
	if (vulkanapi::vkEnumeratePhysicalDevices(vulkan_instance, &physicalDeviceCount, 0) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem enumerating VULKAN physical devices.\n";
		return nullptr;
	}

	//VkPhysicalDevice* const physicalDevices = (VkPhysicalDevice*)malloc(
	//	sizeof(VkPhysicalDevice) * physicalDeviceCount);
	//std::vector<VkPhysicalDevice> physicalDevices(physicalDeviceCount);
	VkPhysicalDevice physicalDevices[16];

	if (vulkanapi::vkEnumeratePhysicalDevices(vulkan_instance, &physicalDeviceCount, &physicalDevices[0]) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem enumerating VULKAN physical devices.\n";
		return nullptr;
	}

	std::cout << "\n" << physicalDeviceCount << " physical devices supported\n";


	VkPhysicalDeviceProperties deviceProperties;
	VkPhysicalDeviceFeatures deviceFeatures;

	VkPhysicalDeviceFeatures2 deviceFeatures2;
	VkPhysicalDeviceVulkan11Features deviceFeatures11;
	VkPhysicalDeviceVulkan12Features deviceFeatures12;
	VkPhysicalDeviceVulkan13Features deviceFeatures13;

	VkPhysicalDeviceSubgroupProperties subgroupProperties;
	VkPhysicalDeviceProperties2 physicalDeviceProperties;

	std::cout << "\navailable devices:\n";

	//VkPhysicalDevice vulkan_physicalDevice;

	//for (const auto& device : physicalDevices) {
	for(int i = 0; i < physicalDeviceCount; i++)
	{
		const auto& device = physicalDevices[i];

		vulkanapi::vkGetPhysicalDeviceProperties(device, &deviceProperties);
		vulkanapi::vkGetPhysicalDeviceFeatures(device, &deviceFeatures);

		deviceFeatures13.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES;
		deviceFeatures13.pNext = nullptr;

		deviceFeatures12.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES;
		deviceFeatures12.pNext = &deviceFeatures13;

		deviceFeatures11.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_FEATURES;
		deviceFeatures11.pNext = &deviceFeatures12;

		deviceFeatures2.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2;
		deviceFeatures2.pNext = &deviceFeatures11;

		vulkanapi::vkGetPhysicalDeviceFeatures2(device, &deviceFeatures2);

		subgroupProperties.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_PROPERTIES;
		subgroupProperties.pNext = nullptr;

		physicalDeviceProperties.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2;
		physicalDeviceProperties.pNext = &subgroupProperties;

		vulkanapi::vkGetPhysicalDeviceProperties2(device, &physicalDeviceProperties);


		std::cout << '\t' << deviceProperties.deviceName << '\n';
		std::cout << '\t' << "devicetype:" << deviceProperties.deviceType << '\n';
		std::cout << '\t' << "apiVersion:" << deviceProperties.apiVersion << '\n';
		std::cout << '\t' << "maxComputeSharedMemorySize:" << deviceProperties.limits.maxComputeSharedMemorySize << '\n';
		std::cout << '\t' << "maxComputeWorkGroupCount:" << deviceProperties.limits.maxComputeWorkGroupCount << '\n';
		std::cout << '\t' << "maxComputeWorkGroupInvocations:" << deviceProperties.limits.maxComputeWorkGroupInvocations << '\n';
		std::cout << '\t' << "maxComputeWorkGroupSize:" << deviceProperties.limits.maxComputeWorkGroupSize << '\n';
		std::cout << '\t' << "maxDescriptorSetStorageBuffers:" << deviceProperties.limits.maxDescriptorSetStorageBuffers << '\n';
		std::cout << '\t' << "subgroupSize: " << subgroupProperties.subgroupSize;


		if ((deviceProperties.deviceType == VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU) || (deviceProperties.deviceType == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU))
		{
			std::cout << "\n***Found suitable device***\n";
			vulkan_physicalDevice = device;

			break;
		}
	}


	//std::cout << "\nVULKAN: DEBUG: storageBuffer16BitAccess:" << deviceFeatures11.storageBuffer16BitAccess;
	//std::cout << "\nVULKAN: DEBUG: shaderInt8:" << deviceFeatures12.shaderInt8;
	//std::cout << "\nVULKAN: DEBUG: storageBuffer8BitAccess:" << deviceFeatures12.storageBuffer8BitAccess;
	//std::cout << "\nVULKAN: DEBUG: uniformAndStorageBuffer8BitAccess:" << deviceFeatures12.uniformAndStorageBuffer8BitAccess;


	// record subgroup size
	vulkan_SubgroupSize = subgroupProperties.subgroupSize;


#ifdef VERBOSE_VULKAN_DEVICE_EXTENSION_INFO
	std::vector<const char*> device_extensions;

	uint32_t deviceExtensionCount = 0;
	vulkanapi::vkEnumerateDeviceExtensionProperties(vulkan_physicalDevice, nullptr, &deviceExtensionCount, nullptr);

	std::cout << "\n" << deviceExtensionCount << " device extensions supported\n";

	std::vector<VkExtensionProperties> device_extensions_props(deviceExtensionCount);

	vulkanapi::vkEnumerateDeviceExtensionProperties(vulkan_physicalDevice, nullptr, &deviceExtensionCount, &device_extensions_props[0]);

	std::cout << "available extensions:\n";
	for (const auto& extension : device_extensions_props) {
		std::cout << '\t' << extension.extensionName << '\n';

		device_extensions.push_back(extension.extensionName);
	}

	//std::cout << "enabling extensions:\n";
	//for (const auto& extension : device_extensions) {
	//	std::cout << '\t' << extension << '\n';
	//}
#endif



	uint32_t queueFamilyPropertiesCount = 0;
	vulkanapi::vkGetPhysicalDeviceQueueFamilyProperties(vulkan_physicalDevice, &queueFamilyPropertiesCount, 0);

	//VkQueueFamilyProperties* const queueFamilyProperties = (VkQueueFamilyProperties*)_alloca(
	//	sizeof(VkQueueFamilyProperties) * queueFamilyPropertiesCount);
	//std::vector<VkQueueFamilyProperties> queueFamilyProperties(queueFamilyPropertiesCount);
	VkQueueFamilyProperties queueFamilyProperties[128];

	vulkanapi::vkGetPhysicalDeviceQueueFamilyProperties(vulkan_physicalDevice, &queueFamilyPropertiesCount, &queueFamilyProperties[0]);

	//uint32_t queueFamilyIndex;
	VkBool32 queue_present_support;


#ifdef VERBOSE_VULKAN_QUEUE_INFO
	// output the queue info
	std::cout << "\nVULKAN: INFO: *** AVAILABLE QUEUES ***";
	for (int i = 0; i < queueFamilyPropertiesCount; i++)
	{
		const VkQueueFlags maskedFlags = queueFamilyProperties[i].queueFlags;

		// need to know if presentation is supported
		vulkanapi::vkGetPhysicalDeviceSurfaceSupportKHR(vulkan_physicalDevice, i, vulkan_presentation_surface, &queue_present_support);

		std::cout << "\n\tQUEUE#" << dec << i
			<< "\n\tSupported queue usages include:"
			<< (maskedFlags & VK_QUEUE_GRAPHICS_BIT ? "\n\t\tVK_QUEUE_GRAPHICS_BIT" : "")
			<< (maskedFlags & VK_QUEUE_COMPUTE_BIT ? "\n\t\tVK_QUEUE_COMPUTE_BIT" : "")
			<< (maskedFlags & VK_QUEUE_TRANSFER_BIT ? "\n\t\tVK_QUEUE_TRANSFER_BIT" : "")
			<< (maskedFlags & VK_QUEUE_SPARSE_BINDING_BIT ? "\n\t\tVK_QUEUE_SPARSE_BINDING_BIT" : "")
			<< (maskedFlags & VK_QUEUE_PROTECTED_BIT ? "\n\t\tVK_QUEUE_PROTECTED_BIT" : "")
			<< (queue_present_support ? "\n\t\t*QUEUE PRESENT SUPPORT*" : "")
			<< std::endl;
	}
#endif


	//for (const auto& properties : queueFamilyProperties) {
	for (int i = 0; i < queueFamilyPropertiesCount; i++)
	{
		// mask out the sparse binding bit that we aren't caring about (yet!) and the transfer bit
		//const VkQueueFlags maskedFlags = (~(VK_QUEUE_TRANSFER_BIT | VK_QUEUE_SPARSE_BINDING_BIT) & queueFamilyProperties[i].queueFlags);
		const VkQueueFlags maskedFlags = queueFamilyProperties[i].queueFlags;

		//if (!(VK_QUEUE_GRAPHICS_BIT & maskedFlags) && (VK_QUEUE_COMPUTE_BIT & maskedFlags)) {
		if (
			(VK_QUEUE_COMPUTE_BIT & maskedFlags)

			//&& (VK_QUEUE_TRANSFER_BIT & maskedFlags)

			// can toggle/modify this for testing
			&& (VK_QUEUE_GRAPHICS_BIT & maskedFlags)
			)
		{

			// this is the queueFamilyIndex for the compute queue, the one for the graphics queue (if there is one) could be different
			vulkan_queueFamilyIndex = i;

			// need to know if presentation is supported
			vulkanapi::vkGetPhysicalDeviceSurfaceSupportKHR(vulkan_physicalDevice, i, vulkan_presentation_surface, &queue_present_support);

			std::cout << "\n***Found suitable queue family index for COMPUTE queue***\n";

			// choose this one right away if it also has presentation support
			if (queue_present_support)
			{
				std::cout << "\n***Found suitable queue family index for COMPUTE queue with PRESENTATION support***\n";
				break;
			}
		}
	}

	VkDeviceQueueCreateInfo deviceQueueCreateInfo{};

	float queuePriority = 1.0f;
	deviceQueueCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
	deviceQueueCreateInfo.queueFamilyIndex = vulkan_queueFamilyIndex;
	deviceQueueCreateInfo.queueCount = 1;
	deviceQueueCreateInfo.pQueuePriorities = &queuePriority;
	deviceQueueCreateInfo.flags = 0;
	deviceQueueCreateInfo.pNext = 0;


	const char* required_device_extensions[] = {
	  VK_KHR_SWAPCHAIN_EXTENSION_NAME
	};


	VkDeviceCreateInfo deviceCreateInfo{};

	deviceCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
	deviceCreateInfo.pNext = &deviceFeatures11;

	deviceCreateInfo.pQueueCreateInfos = &deviceQueueCreateInfo;
	deviceCreateInfo.queueCreateInfoCount = 1;

	// enable all the device features
	deviceCreateInfo.pEnabledFeatures = &deviceFeatures;
	//deviceCreateInfo.pEnabledFeatures = 0;


	deviceCreateInfo.enabledExtensionCount = 1;
	deviceCreateInfo.ppEnabledExtensionNames = &required_device_extensions[0];
	//deviceCreateInfo.enabledExtensionCount = static_cast<uint32_t>(device_extensions.size());
	//deviceCreateInfo.ppEnabledExtensionNames = device_extensions.data();
	//deviceCreateInfo.enabledExtensionCount = 0;
	//deviceCreateInfo.ppEnabledExtensionNames = nullptr;

	deviceCreateInfo.enabledLayerCount = 0;
	deviceCreateInfo.flags = 0;


	//VkDevice vulkan_device;
	if (vulkanapi::vkCreateDevice(vulkan_physicalDevice, &deviceCreateInfo, nullptr, &vulkan_device) != VK_SUCCESS) {
		//throw std::runtime_error("failed to create logical device!");
		std::cout << "\nVULKAN ERROR: Problem creating VULKAN device.\n";
		return nullptr;
	}
	else
	{
		std::cout << "\nVULKAN SUCCESS: Successfully created VULKAN device.\n";

	}



	if (!vulkanapi::vulkan_load_device_level_entry_points())
	{
		std::cout << "\nERROR: VULKAN: INIT: Unable to load device level entry points!\n";
		return nullptr;

	}








	// create command pool ??
	VkCommandPoolCreateInfo commandPoolCreateInfo = {
	  VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO,
	  0,
	  VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT,
	  vulkan_queueFamilyIndex
	};

	//VkCommandPool vulkan_commandPool;
	if (vulkanapi::vkCreateCommandPool(vulkan_device, &commandPoolCreateInfo, 0, &vulkan_commandPool) != VK_SUCCESS)
	{
		cout << "\nVULKAN ERROR: Problem creating command pool.\n";
		return false;
	}


	VkQueue computeQueue;
	vulkanapi::vkGetDeviceQueue(vulkan_device, vulkan_queueFamilyIndex, 0, &computeQueue);

	// *** device created *** //


	// create semaphores
	VkSemaphoreCreateInfo semaphore_create_info = {
	  VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,      // VkStructureType          sType
	  nullptr,                                      // const void*              pNext
	  0                                             // VkSemaphoreCreateFlags   flags
	};

	//VkSemaphore vulkan_image_available_semaphore;
	//VkSemaphore vulkan_rendering_finished_semaphore;
	if ((vulkanapi::vkCreateSemaphore(vulkan_device, &semaphore_create_info, nullptr, &vulkan_image_available_semaphore) != VK_SUCCESS) ||
		(vulkanapi::vkCreateSemaphore(vulkan_device, &semaphore_create_info, nullptr, &vulkan_rendering_finished_semaphore) != VK_SUCCESS)) {

		std::cout << "ERROR: VULKAN: Could not create semaphores!" << std::endl;
		return nullptr;
	}



	// set gpu type integrated or discrete
	vulkan_device_type = deviceProperties.deviceType;



	VkPhysicalDeviceMemoryProperties properties;

	vulkanapi::vkGetPhysicalDeviceMemoryProperties(vulkan_physicalDevice, &properties);

	
	VkImageCreateInfo vulkan_image_info =
	{
		VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO,
		nullptr,
		0,
		VK_IMAGE_TYPE_2D,

		// I want a 24-bit rgb format, but TODO - need to check to make sure format is supported
		//VK_FORMAT_R32G32B32A32_SFLOAT,	//VK_FORMAT_R8G8B8A8_UINT,
		VK_FORMAT_R8G8B8A8_UNORM,

		// size of the image
		{ VULKAN_IMAGE_WIDTH, VULKAN_IMAGE_HEIGHT, 1 },

		1, 1,
		VK_SAMPLE_COUNT_1_BIT,
		//VK_IMAGE_TILING_LINEAR,
		VK_IMAGE_TILING_OPTIMAL,

		// just storing image for use in compute shader for now
		VK_IMAGE_USAGE_STORAGE_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT | VK_IMAGE_USAGE_TRANSFER_DST_BIT,

		// only using compute queue for now
		VK_SHARING_MODE_EXCLUSIVE,

		// this is just going to be on the compute queue for now, TODO might also need to add graphics queue later for displaying onscreen?
		1, &vulkan_queueFamilyIndex,

		// validation layer: vulkan specification says that cannot init image with VK_IMAGE_LAYOUT_GENERAL
		VK_IMAGE_LAYOUT_UNDEFINED
	};


	//VkImage vulkan_image;
	if (vulkanapi::vkCreateImage(vulkan_device, &vulkan_image_info, nullptr, &vulkan_image) != VK_SUCCESS)
	{
		cout << "\nERROR: VULKAN: problem creating image.";
	}

	VkMemoryRequirements image_memReqs;
	vulkanapi::vkGetImageMemoryRequirements(vulkan_device, vulkan_image, &image_memReqs);
	

#ifdef VERBOSE_VULKAN_MEMORY_INFO
	for (int i = 0; i < properties.memoryTypeCount; i++)
	{
		const VkMemoryPropertyFlags maskedFlags = properties.memoryTypes[i].propertyFlags;

		std::cout << "\n\tMEMORY INDEX#" << dec << i
			<< "\n\tMemory Size:" << dec << properties.memoryHeaps[properties.memoryTypes[i].heapIndex].size
			<< "\n\tSupported memory usages include:"
			<< (maskedFlags & VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT ? "\n\t\tVK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT ? "\n\t\tVK_MEMORY_PROPERTY_HOST_VISIBLE_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_HOST_COHERENT_BIT ? "\n\t\tVK_MEMORY_PROPERTY_HOST_COHERENT_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_HOST_CACHED_BIT ? "\n\t\tVK_MEMORY_PROPERTY_HOST_CACHED_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_LAZILY_ALLOCATED_BIT ? "\n\t\tVK_MEMORY_PROPERTY_LAZILY_ALLOCATED_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_PROTECTED_BIT ? "\n\t\tVK_MEMORY_PROPERTY_PROTECTED_BIT" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_DEVICE_COHERENT_BIT_AMD ? "\n\t\tVK_MEMORY_PROPERTY_DEVICE_COHERENT_BIT_AMD" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_DEVICE_UNCACHED_BIT_AMD ? "\n\t\tVK_MEMORY_PROPERTY_DEVICE_UNCACHED_BIT_AMD" : "")
			<< (maskedFlags & VK_MEMORY_PROPERTY_RDMA_CAPABLE_BIT_NV ? "\n\t\tVK_MEMORY_PROPERTY_RDMA_CAPABLE_BIT_NV" : "")
			<< std::endl;
	}
#endif


	int total_vars_count = 0;
	//std::vector<VkDeviceSize> vulkan_bufferSizes(external_vars_count + internal_vars_count);

	// get size of external memory buffers //

	vulkan_external_vars_count = external_vars_count;

	// add up external memory total size
	VkDeviceSize memorySize = 0;
	for (int i = 0; i < external_vars_count; i++)
	{
		vulkan_bufferSizes[total_vars_count++] = external_vars_sizes[i];
		memorySize += external_vars_sizes[i];
	}


	// get size of local memory buffers //

	// note: if discrete gpu, then this includes external buffers

	VkDeviceSize memorySize_local = 0;

	// add up internal memory total size
	for (int i = 0; i < internal_vars_count; i++)
	{
		vulkan_bufferSizes[total_vars_count++] = internal_vars_sizes[i];
		memorySize_local += internal_vars_sizes[i];
	}

	// get total count of vars
	//internal_vars_count += external_vars_count;


	// show total number of vars
	std::cout << "\nVULKAN: INFO: total number of vars: " << dec << total_vars_count;
	std::cout << "\n\tnumber of external vars: " << dec << external_vars_count;
	std::cout << "\n\tnumber of internal vars: " << dec << internal_vars_count;


	// output memory types availables


	// set memoryTypeIndex to an invalid entry in the properties.memoryTypes array
	uint32_t memoryTypeIndex = VK_MAX_MEMORY_TYPES;
	if (vulkan_device_type == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU)
	{
		for (uint32_t k = 0; k < properties.memoryTypeCount; k++)
		{
			if (
				(VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_COHERENT_BIT & properties.memoryTypes[k].propertyFlags) &&
				//(VK_MEMORY_PROPERTY_HOST_CACHED_BIT & properties.memoryTypes[k].propertyFlags) &&
				(memorySize < properties.memoryHeaps[properties.memoryTypes[k].heapIndex].size)) {
				memoryTypeIndex = k;
				break;
			}
		}
	}
	else
	{
		for (uint32_t k = 0; k < properties.memoryTypeCount; k++)
		{
			if (
				//(VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_COHERENT_BIT & properties.memoryTypes[k].propertyFlags) &&
				(VK_MEMORY_PROPERTY_HOST_CACHED_BIT & properties.memoryTypes[k].propertyFlags) &&
				(memorySize < properties.memoryHeaps[properties.memoryTypes[k].heapIndex].size)) {
				memoryTypeIndex = k;
				break;
			}
		}
	}

	if (memoryTypeIndex == VK_MAX_MEMORY_TYPES)
	{
		std::cout << "\nERROR: VULKAN: Problem selecting external memory type.\n";
		return nullptr;
	}


	uint32_t memoryTypeIndex_local = VK_MAX_MEMORY_TYPES;
	for (uint32_t k = 0; k < properties.memoryTypeCount; k++) {
		if (
			(VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT & properties.memoryTypes[k].propertyFlags) &&
			//(VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT & properties.memoryTypes[k].propertyFlags) &&
			//(VK_MEMORY_PROPERTY_HOST_COHERENT_BIT & properties.memoryTypes[k].propertyFlags) &&
			//(VK_MEMORY_PROPERTY_HOST_CACHED_BIT & properties.memoryTypes[k].propertyFlags) &&
			(memorySize_local < properties.memoryHeaps[properties.memoryTypes[k].heapIndex].size)) {
			memoryTypeIndex_local = k;
			break;
		}
	}

	if (memoryTypeIndex_local == VK_MAX_MEMORY_TYPES)
	{
		std::cout << "\nERROR: VULKAN: Problem selecting local memory type.\n";
		return nullptr;
	}



	cout << "\nEXTERNAL memoryTypeIndex=" << dec << memoryTypeIndex << " memoryTypeCount=" << properties.memoryTypeCount << " VK_MAX_MEMORY_TYPES=" << VK_MAX_MEMORY_TYPES;
	cout << "\nLOCAL memoryTypeIndex=" << dec << memoryTypeIndex_local << " memoryTypeCount=" << properties.memoryTypeCount << " VK_MAX_MEMORY_TYPES=" << VK_MAX_MEMORY_TYPES;


	// allocate external memory //


	const VkMemoryAllocateInfo memoryAllocateInfo = {
		  VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
		  0,
		  memorySize,
		  memoryTypeIndex
	};

	//VkDeviceMemory vulkan_memory;
	if (vulkanapi::vkAllocateMemory(vulkan_device, &memoryAllocateInfo, 0, &vulkan_memory) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating memory for device.\n";
		return nullptr;

	}


	// allocate local memory //

	// note: if discrete gpu, then this includes external vars

	const VkMemoryAllocateInfo memoryAllocateInfo_local = {
		  VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
		  0,
		  memorySize_local,
		  memoryTypeIndex_local
	};

	//VkDeviceMemory vulkan_memory;
	if (vulkanapi::vkAllocateMemory(vulkan_device, &memoryAllocateInfo_local, 0, &vulkan_memory_local) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating local memory for device.\n";
		return nullptr;

	}


	// allocate memory for image //


	const VkMemoryAllocateInfo memoryAllocateInfo_image = {
		  VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
		  0,
		  image_memReqs.size,
		  memoryTypeIndex_local
	};

	//VkDeviceMemory vulkan_memory;
	if (vulkanapi::vkAllocateMemory(vulkan_device, &memoryAllocateInfo_image, 0, &vulkan_memory_image) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating local image memory for device.\n";
		return nullptr;

	}

	vulkanapi::vkBindImageMemory(vulkan_device, vulkan_image, vulkan_memory_image, 0);


	int current_offset = 0;

	//vulkan_bufferCreateInfos.resize(total_vars_count);
	//vulkan_buffers.resize(total_vars_count);
	vulkan_buffers_count = total_vars_count;

	//std::vector<VkBufferCreateInfo> vulkan_bufferCreateInfos(total_vars_count);
	//std::vector<VkDescriptorSetLayoutBinding> descriptorSetLayoutBindings(total_vars_count);
	VkBufferCreateInfo vulkan_bufferCreateInfos[32];
	VkDescriptorSetLayoutBinding descriptorSetLayoutBindings[32];
	uint32_t descriptorSetLayoutBindings_count = 0;

	for (uint32_t i = 0; i < external_vars_count; i++)
	{
		//std::vector<VkBufferCreateInfo> vulkan_bufferCreateInfos(0);
		//std::vector<VkBuffer> vulkan_buffers(0);

		vulkan_bufferCreateInfos[i] = {
		  VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
		  0,
		  0,
		  vulkan_bufferSizes[i],
		  VK_BUFFER_USAGE_STORAGE_BUFFER_BIT | VK_BUFFER_USAGE_TRANSFER_SRC_BIT | VK_BUFFER_USAGE_TRANSFER_DST_BIT,
		  VK_SHARING_MODE_EXCLUSIVE,
		  1,
		  &vulkan_queueFamilyIndex
		};

		if (vulkanapi::vkCreateBuffer(vulkan_device, &vulkan_bufferCreateInfos[i], 0, &vulkan_buffers[i]) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem creating input buffer.\n";
			return nullptr;

		}

		//if (vulkanapi::vkBindBufferMemory(vulkan_device, vulkan_buffers[i], vulkan_memory, 0) != VK_SUCCESS)
		if (vulkanapi::vkBindBufferMemory(vulkan_device, vulkan_buffers[i], vulkan_memory, current_offset) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem binding internal vram buffer.\n";
			return nullptr;

		}

		current_offset += vulkan_bufferSizes[i];

		descriptorSetLayoutBindings[descriptorSetLayoutBindings_count] = { descriptorSetLayoutBindings_count, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 1, VK_SHADER_STAGE_COMPUTE_BIT, 0 };
		descriptorSetLayoutBindings_count++;
	}


	current_offset = 0;
	for (uint32_t i = external_vars_count; i < total_vars_count; i++)
	{
		//std::vector<VkBufferCreateInfo> vulkan_bufferCreateInfos(0);
		//std::vector<VkBuffer> vulkan_buffers(0);

		vulkan_bufferCreateInfos[i] = {
		  VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
		  0,
		  0,
		  vulkan_bufferSizes[i],
		  VK_BUFFER_USAGE_STORAGE_BUFFER_BIT | VK_BUFFER_USAGE_TRANSFER_SRC_BIT | VK_BUFFER_USAGE_TRANSFER_DST_BIT,
		  VK_SHARING_MODE_EXCLUSIVE,
		  1,
		  &vulkan_queueFamilyIndex
		};

		if (vulkanapi::vkCreateBuffer(vulkan_device, &vulkan_bufferCreateInfos[i], 0, &vulkan_buffers[i]) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem creating input buffer.\n";
			return nullptr;

		}

		//if (vulkanapi::vkBindBufferMemory(vulkan_device, vulkan_buffers[i], vulkan_memory, 0) != VK_SUCCESS)
		if (vulkanapi::vkBindBufferMemory(vulkan_device, vulkan_buffers[i], vulkan_memory_local, current_offset) != VK_SUCCESS)
		{
			std::cout << "\nVULKAN ERROR: Problem binding internal vram buffer.\n";
			return nullptr;

		}

		current_offset += vulkan_bufferSizes[i];

		descriptorSetLayoutBindings[descriptorSetLayoutBindings_count] = { descriptorSetLayoutBindings_count, VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 1, VK_SHADER_STAGE_COMPUTE_BIT, 0 };
		descriptorSetLayoutBindings_count++;
	}



	int32_t* payload;



	std::cout << "\nVULKAN: INFO: Creating buffers and binding memory.\n";




	std::cout << "\nVULKAN: INFO: loading and creating compute shader.\n";


	// *** load shader here *** //

	VkShaderModuleCreateInfo shaderModuleCreateInfo = {
		  VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
		  0,
		  0,
		(size_t)shader_file_size,
		reinterpret_cast<const uint32_t*>(shader_file_data)
	};


	//VkShaderModule vulkan_shader_module;

	if (vulkanapi::vkCreateShaderModule(vulkan_device, &shaderModuleCreateInfo, 0, &vulkan_shader_module) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating shader module.\n";
		return nullptr;

	}


	std::cout << "\nVULKAN: INFO: creating layouts.\n";




	VkDescriptorSetLayoutCreateInfo descriptorSetLayoutCreateInfo = {
	  VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO,
	  //0, 0, 7, descriptorSetLayoutBindings
	  0, 0, descriptorSetLayoutBindings_count, &descriptorSetLayoutBindings[0]
	};


	//VkDescriptorSetLayout vulkan_descriptorSetLayout;
	if (vulkanapi::vkCreateDescriptorSetLayout(vulkan_device, &descriptorSetLayoutCreateInfo, 0, &vulkan_descriptorSetLayout) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating descriptor set layout.\n";
		return nullptr;

	}


	VkPipelineLayoutCreateInfo pipelineLayoutCreateInfo = {
	  VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO,
	  0,
	  0,
	  1,
	  &vulkan_descriptorSetLayout,
	  0,
	  0
	};

	//VkPipelineLayout vulkan_pipelineLayout;
	if (vulkanapi::vkCreatePipelineLayout(vulkan_device, &pipelineLayoutCreateInfo, 0, &vulkan_pipelineLayout) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating pipeline layout.\n";
		return nullptr;

	}

	VkComputePipelineCreateInfo computePipelineCreateInfo = {
	  VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO,
	  nullptr, 0,
	  {
		VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
		nullptr, 0, VK_SHADER_STAGE_COMPUTE_BIT, vulkan_shader_module, "main", nullptr
	  },
	  vulkan_pipelineLayout, VK_NULL_HANDLE, 0
	};

	//VkPipeline vulkan_pipeline;
	VkResult vkres;
	vkres = vulkanapi::vkCreateComputePipelines(vulkan_device, VK_NULL_HANDLE, 1, &computePipelineCreateInfo, 0, &vulkan_pipeline);
	//if (vkCreateComputePipelines(vulkan_device, 0, 1, &computePipelineCreateInfo, 0, &vulkan_pipeline) != VK_SUCCESS)
	if (vkres != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating compute pipelines.\n";
		std::cout << "\nvkres=" << vkres;
		return nullptr;

	}



	VkDescriptorPoolSize descriptorPoolSize[1] = {
		{
			VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
			descriptorSetLayoutBindings_count	// 7
		}
	};

	//VkDescriptorPoolSize descriptorPoolSize = {
	// VK_DESCRIPTOR_TYPE_STORAGE_IMAGE,
	 // 1
	//};

	VkDescriptorPoolCreateInfo descriptorPoolCreateInfo = {
	  VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO,
	  0,
	  0,
	  1,	// 1,
	  1,	// 1,
	  descriptorPoolSize
	};

	//VkDescriptorPool vulkan_descriptorPool;
	if (vulkanapi::vkCreateDescriptorPool(vulkan_device, &descriptorPoolCreateInfo, 0, &vulkan_descriptorPool) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem creating descriptor pool.\n";
		return nullptr;

	}

	VkDescriptorSetAllocateInfo descriptorSetAllocateInfo = {
	  VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO,
	  0,
	  vulkan_descriptorPool,
	  1,
	  &vulkan_descriptorSetLayout
	};

	//VkDescriptorSet vulkan_descriptorSet;
	if (vulkanapi::vkAllocateDescriptorSets(vulkan_device, &descriptorSetAllocateInfo, &vulkan_descriptorSet) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem allocating descriptor sets.\n";
		return nullptr;

	}


	//std::vector<VkDescriptorBufferInfo> descriptorBufferInfo(total_vars_count);
	//std::vector<VkWriteDescriptorSet> writeDescriptorSet(total_vars_count);
	VkDescriptorBufferInfo descriptorBufferInfo[32];
	VkWriteDescriptorSet writeDescriptorSet[32];

	uint32_t descriptor_start_idx = 0;

	uint32_t writeDescriptorSet_count = 0;
	for (uint32_t i = descriptor_start_idx; i < total_vars_count; i++)
	{
		descriptorBufferInfo[writeDescriptorSet_count] = {
			vulkan_buffers[i],
			0,
			VK_WHOLE_SIZE
		};

		writeDescriptorSet[writeDescriptorSet_count] = {
			VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
			0,
			vulkan_descriptorSet,
			writeDescriptorSet_count,
			0,
			1,
			VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
			nullptr,
			&descriptorBufferInfo[writeDescriptorSet_count],
			nullptr
		};

		writeDescriptorSet_count++;
	}




	std::cout << "\nVULKAN: INFO: creating descriptor set.\n";




	//vulkanapi::vkUpdateDescriptorSets(vulkan_device, 7, writeDescriptorSet, 0, 0);
	vulkanapi::vkUpdateDescriptorSets(vulkan_device, writeDescriptorSet_count, &writeDescriptorSet[0], 0, 0);




	



	//VkSubmitInfo submitInfo;
	vulkan_submitInfo = {
	  VK_STRUCTURE_TYPE_SUBMIT_INFO,
	  0,
	  0,
	  0,
	  0,
	  1,
	  &vulkan_commandBuffer1,
	  0,
	  0
	};


	//VkQueue queue;
	vulkanapi::vkGetDeviceQueue(vulkan_device, vulkan_queueFamilyIndex, 0, &vulkan_queue);

	// create fence
	VkFenceCreateInfo fenceInfo{};
	fenceInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
	fenceInfo.flags = VK_FENCE_CREATE_SIGNALED_BIT;

	if (vulkanapi::vkCreateFence(vulkan_device, &fenceInfo, nullptr, &vulkan_fence) != VK_SUCCESS) {
		std::cout << "\nVULKAN: ERROR: failed to create fence!";
		return nullptr;
	}


	std::cout << "\nVULKAN: INFO: mapping memory for compute shader\n";

	//if (vkMapMemory(vulkan_device, vulkan_memory, 0, memorySize - work_buffer_size_bytes - image_memReqs.size, 0, (void**)&payload) != VK_SUCCESS)
	if (vulkanapi::vkMapMemory(vulkan_device, vulkan_memory, 0, memorySize, 0, (void**)&payload) != VK_SUCCESS)
	{
		std::cout << "\nVULKAN ERROR: Problem mapping memory for device.\n";
		return nullptr;

	}


	// set the mapped memory
	vulkan_mappedmemory = payload;


	std::cout << "\nVULKAN: INFO: creating swap chain.\n";

	// create swap chain
	if (!vulkan_create_swap_chain())
	{
		std::cout << "\nVULKAN: SETUP: ERROR: problem creating SWAP CHAIN.";
		return nullptr;
	}



	// record the default command buffers to use
	//if (!vulkan_record_default_commands())
	//{
	//	std::cout << "\nVULKAN: ERROR: Problem recording command buffers.\n";
	//	return false;
	//}


	// show device details
	cout << "\n*** VULKAN SETUP SUCCESSFUL ***";
	cout << "\nDevice Name: " << deviceProperties.deviceName;
	cout << "\nDevice Type: " << ((deviceProperties.deviceType == 1) ? "INTEGRATED-GPU" : "DISCRETE-GPU");
	cout << "\nDevice local subgroup size: " << dec << vulkan_SubgroupSize;
	//cout << "\nDevice number of local subgroups: " << dec << vulkan_NumSubgroups;
	//cout << "\nDevice local workgroup size: " << dec << vulkan_WorkgroupSize;
	//cout << "\nDevice number of total workgroups: " << dec << vulkan_NumWorkgroups;



	// vulkan loaded
	vulkan_loaded = true;


	std::cout << "\n***VULKAN SETUP COMPLETED SUCCESSFULLY***.\n";
	std::cout << "\nvulkan alloc count: " << vulkan_alloc_count;
	std::cout << "\nvulkan alloc size: " << vulkan_alloc_size;

	// return mapped memory for device
	return (void*)payload;
}






bool vulkan_is_loaded()
{
	return (vulkan_loaded != 0);
}

