/*
    Copyright (C) 2012-2030

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

//#pragma once
#ifndef __VULKAN_SETUP_H__
#define __VULKAN_SETUP_H__

#include "WinApiHandler.h"


extern uint32_t vulkan_new_job_submit;

/// <summary>
/// loads vulkan library
/// returns false when there was a problem
/// return true when successful
/// </summary>
/// <returns>bool</returns>
bool vulkan_load_library();

namespace vulkanapi {

	bool vulkan_load_exported_entry_points();
	bool vulkan_load_global_level_entry_points();
	bool vulkan_load_instance_level_entry_points();
	bool vulkan_load_device_level_entry_points();

}


// need to know if vulkan is already loaded
bool vulkan_is_loaded();



// must do this before setup - sets size of window to output to
void vulkan_set_screen_size(int screen_size_x, int screen_size_y);


// set the size of the buffer
void vulkan_set_buffer_size(int buffer_size_x, int buffer_size_y);

// need to know if the screen or buffer size has changed
bool vulkan_is_screen_size_changed(int screen_size_x, int screen_size_y);
bool vulkan_is_buffer_size_changed(int buffer_size_x, int buffer_size_y);

// also must do this before setup or before recreating swap chain if it changes
void vulkan_set_gpu_threads(int num_gpu_threads);
int vulkan_get_gpu_threads();


// need to know subgroup size
int vulkan_get_subgroup_size();

// create a vulkan context for the window
bool vulkan_create_context(HWND wHandle, HINSTANCE wInstance);

// create/re-create swap chain
bool vulkan_create_swap_chain();

// setup vulkan compute
void* vulkan_setup_compute(char* shader_file_data, int shader_file_size, long* external_vars_sizes, int external_vars_count, long* internal_vars_sizes, int internal_vars_count);

// setup vulkan
// shader_file_name - name of the file that contains the computer shader
// inout_buffer_size_bytes - size of ps1/ps2 gpu ram (1MB or 4MB, etc)
// out_buffer_size_bytes - size of ps1/ps2 screen output
// in_comm_buffer_size_bytes - size of input draw commands buffer
// in_data_buffer_size_bytes - size of input data buffer (graphics data)
// work_buffer_size_bytes - size of work buffer for rasterizing primitives (number of input command buffer slots times data needed per slot)
//void* vulkan_setup(HWND wHandle, HINSTANCE wInstance, char* shader_file_data, int shader_file_size, int inout_buffer_size_bytes, int out_buffer_size_bytes, int in_comm_buffer_size_bytes, int in_data_buffer_size_bytes, int work_buffer_size_bytes);
void* vulkan_setup(HWND wHandle, HINSTANCE wInstance, char* shader_file_data, int shader_file_size, long* external_vars_sizes, int external_vars_count, long* internal_vars_sizes, int internal_vars_count );




// check if commands are recorded //

// check if the commands needed for running ps2 gpu on compute shader and displaying are recorded
bool vulkan_is_compute_draw_recorded();

// check if the commands for only sending the rendered graphics to the screen are recorded
bool vulkan_is_draw_only_recorded();



// record the commands //

// record default commands (for running ps2 gpu on hardware and displaying) using swap chain
bool vulkan_record_default_commands();

// record commands for only displaying buffer to screen using swap chain
bool vulkan_record_draw_only_commands();


// execute vulcan job (compute shader run, possibly later update screen, etc)
bool vulkan_execute();
bool vulkan_execute_compute_only();
bool vulkan_execute_draw_only();

bool vulkan_execute_display(int global_barrier_count, bool draw_to_screen = true);
bool vulkan_record_execute_nodisplay();


bool vulkan_prerecord_command_buffers(int max_thread_count);



// wait for last vulkan job to finish
// returns true if successful, returns false on error
bool vulkan_wait();


// wait for fence submitted with last job to signal
// returns true if successful, returns false on error
bool vulkan_wait_fence();

// reset fence
bool vulkan_reset_fence();


// flush mapped memory writes
void vulkan_flush();

// release vulkan resources
void vulkan_destroy();

#endif  // end #ifndef __VULKAN_SETUP_H__
